"""All Dataset related functions
"""
from typing import List, Optional

from relevanceai.client.helpers import Credentials
from relevanceai.utils.base import _Base
from relevanceai._api.endpoints.datasets.documents import DocumentsClient
from relevanceai._api.endpoints.datasets.monitor import MonitorClient
from relevanceai._api.endpoints.datasets.tasks import TasksClient
from relevanceai._api.endpoints.datasets.cluster import ClusterClient


class DatasetsClient(_Base):
    """All dataset-related functions"""

    def __init__(self, credentials: Credentials):
        self.tasks = TasksClient(credentials)
        self.documents = DocumentsClient(credentials)
        self.monitor = MonitorClient(credentials)
        self.cluster = ClusterClient(credentials)

        super().__init__(credentials)

    def schema(self, dataset_id: str):
        """
        Returns the schema of a dataset. Refer to datasets.create for different field types available in a VecDB schema.

        Parameters
        ----------
        dataset_id : string
            Unique name of dataset
        """
        return self.make_http_request(
            endpoint=f"/datasets/{dataset_id}/schema", method="GET"
        )

    def metadata(self, dataset_id: str):
        """
        Retreives metadata about a dataset. Notably description, data source, etc

        Parameters
        ----------
        dataset_id : string
            Unique name of dataset
        """
        return self.make_http_request(
            endpoint=f"/datasets/{dataset_id}/metadata", method="GET"
        )

    def post_metadata(self, dataset_id: str, metadata: dict):
        """
        Edit and add metadata about a dataset. Notably description, data source, etc
        """
        return self.make_http_request(
            endpoint=f"/datasets/{dataset_id}/metadata",
            method="POST",
            parameters={"dataset_id": dataset_id, "metadata": metadata},
        )

    def create(self, dataset_id: str, schema: Optional[dict] = None):
        """
        A dataset can store documents to be searched, retrieved, filtered and aggregated (similar to Collections in MongoDB, Tables in SQL, Indexes in ElasticSearch).
        A powerful and core feature of VecDB is that you can store both your metadata and vectors in the same document. When specifying the schema of a dataset and inserting your own vector use the suffix (ends with) "_vector_" for the field name, and specify the length of the vector in dataset_schema. \n

        For example:

        >>>    {
        >>>        "product_image_vector_": 1024,
        >>>        "product_text_description_vector_" : 128
        >>>    }

        These are the field types supported in our datasets: ["text", "numeric", "date", "dict", "chunks", "vector", "chunkvector"]. \n

        For example:

        >>>    {
        >>>        "product_text_description" : "text",
        >>>        "price" : "numeric",
        >>>        "created_date" : "date",
        >>>        "product_texts_chunk_": "chunks",
        >>>        "product_text_chunkvector_" : 1024
        >>>    }

        You don't have to specify the schema of every single field when creating a dataset, as VecDB will automatically detect the appropriate data type for each field (vectors will be automatically identified by its "_vector_" suffix). Infact you also don't always have to use this endpoint to create a dataset as /datasets/bulk_insert will infer and create the dataset and schema as you insert new documents. \n

        Note:

            - A dataset name/id can only contain undercase letters, dash, underscore and numbers.
            - "_id" is reserved as the key and id of a document.
            - Once a schema is set for a dataset it cannot be altered. If it has to be altered, utlise the copy dataset endpoint.

        For more information about vectors check out the 'Vectorizing' section, services.search.vector or out blog at https://relevance.ai/blog. For more information about chunks and chunk vectors check out services.search.chunk.

        Parameters
        ----------
        dataset_id : string
            Unique name of dataset
        schema : dict
            Schema for specifying the field that are vectors and its length

        """
        schema = {} if schema is None else schema

        return self.make_http_request(
            endpoint=f"/datasets/create",
            method="POST",
            parameters={"id": dataset_id, "schema": schema},
        )

    def list(self):
        """List all datasets in a project that you are authorized to read/write."""
        return self.make_http_request(endpoint="/datasets/list", method="GET")

    def list_all(
        self,
        include_schema: bool = True,
        include_stats: bool = True,
        include_metadata: bool = True,
        include_schema_stats: bool = False,
        include_vector_health: bool = False,
        include_active_jobs: bool = False,
        dataset_ids: Optional[list] = None,
        sort_by_created_at_date: bool = False,
        asc: bool = False,
        page_size: int = 20,
        page: int = 1,
    ):
        """
        Returns a page of datasets and in detail the dataset's associated information that you are authorized to read/write. The information includes:

        - Schema - Data schema of a dataset (same as dataset.schema).
        - Metadata - Metadata of a dataset (same as dataset.metadata).
        - Stats - Statistics of number of documents and size of a dataset (same as dataset.stats).
        - Vector_health - Number of zero vectors stored (same as dataset.health).
        - Schema_stats - Fields and number of documents missing/not missing for that field (same as dataset.stats).
        - Active_jobs - All active jobs/tasks on the dataset.

        Parameters
        ----------
        include_schema : bool
            Whether to return schema
        include_stats : bool
            Whether to return stats
        include_metadata : bool
            Whether to return metadata
        include_vector_health : bool
            Whether to return vector_health
        include_schema_stats : bool
            Whether to return schema_stats
        include_active_jobs : bool
            Whether to return active_jobs
        dataset_ids : list
            List of dataset IDs
        sort_by_created_at_date : bool
            Sort by created at date. By default shows the newest datasets. Set asc=False to get oldest dataset.
        asc : bool
            Whether to sort results by ascending or descending order
        page_size : int
            Size of each page of results
        page : int
            Page of the results
        """
        dataset_ids = [] if dataset_ids is None else dataset_ids

        return self.make_http_request(
            endpoint="/datasets/list",
            method="POST",
            parameters={
                "include_schema": include_schema,
                "include_stats": include_stats,
                "include_metadata": include_metadata,
                "include_schema_stats": include_schema_stats,
                "include_vector_health": include_vector_health,
                "include_active_jobs": include_active_jobs,
                "dataset_ids": dataset_ids,
                "sort_by_created_at_date": sort_by_created_at_date,
                "asc": asc,
                "page_size": page_size,
                "page": page,
            },
        )

    def facets(
        self,
        dataset_id,
        fields: Optional[list] = None,
        date_interval: str = "monthly",
        page_size: int = 5,
        page: int = 1,
        asc: bool = False,
    ):
        """
        Takes a high level aggregation of every field, return their unique values and frequencies. This is used to help create the filter bar for search.

        Parameters
        ----------
        dataset_id : string
            Unique name of dataset
        fields : list
            Fields to include in the facets, if [] then all
        date_interval : str
            Interval for date facets
        page_size : int
            Size of facet page
        page : int
            Page of the results
        asc: bool
            Whether to sort results by ascending or descending order

        """
        fields = [] if fields is None else fields

        return self.make_http_request(
            endpoint=f"/datasets/{dataset_id}/facets",
            method="POST",
            parameters={
                "fields": fields,
                "date_interval": date_interval,
                "page_size": page_size,
                "page": page,
                "asc": asc,
            },
        )

    def check_missing_ids(self, dataset_id, ids):

        """
        Look up in bulk if the ids exists in the dataset, returns all the missing one as a list.

        Parameters
        ----------
        dataset_id : string
            Unique name of dataset
        ids : list
            IDs of documents
        """

        # Check if dataset_id exists
        dataset_exists = dataset_id in self.list()["datasets"]

        if dataset_exists:
            return self.make_http_request(
                endpoint=f"/datasets/{dataset_id}/documents/get_missing",
                method="GET",
                parameters={"ids": ids},
            )

        else:
            print("Dataset does not exist")
            return

    def insert(
        self,
        dataset_id: str,
        document: dict,
        insert_date: bool = True,
        overwrite: bool = True,
        update_schema: bool = True,
    ):
        """
        Insert a single documents

        - When inserting the document you can optionally specify your own id for a document by using the field name "_id", if not specified a random id is assigned.
        - When inserting or specifying vectors in a document use the suffix (ends with) "_vector_" for the field name. e.g. "product_description_vector_".
        - When inserting or specifying chunks in a document the suffix (ends with) "_chunk_" for the field name. e.g. "products_chunk_".
        - When inserting or specifying chunk vectors in a document's chunks use the suffix (ends with) "_chunkvector_" for the field name. e.g. "products_chunk_.product_description_chunkvector_".

        Documentation can be found here: https://ingest-api-dev-aueast.relevance.ai/latest/documentation#operation/InsertEncode \n

        Try to keep each batch of documents to insert under 200mb to avoid the insert timing out. \n

        Parameters
        ----------
        dataset_id : string
            Unique name of dataset
        documents : list
            A list of documents. Document is a JSON-like data that we store our metadata and vectors with. For specifying id of the document use the field '_id', for specifying vector field use the suffix of '_vector_'
        insert_date : bool
            Whether to include insert date as a field 'insert_date_'.
        overwrite : bool
            Whether to overwrite document if it exists.
        update_schema : bool
            Whether the api should check the documents for vector datatype to update the schema.

        """

        return self.make_http_request(
            endpoint=f"/datasets/{dataset_id}/documents/insert",
            method="POST",
            parameters={
                "document": document,
                "insert_date": insert_date,
                "overwrite": overwrite,
                "update_schema": update_schema,
            },
        )

    def bulk_insert(
        self,
        dataset_id: str,
        documents: List,
        insert_date: bool = True,
        overwrite: bool = True,
        update_schema: bool = True,
        field_transformers: Optional[list] = None,
        return_documents: bool = False,
    ):
        """
        Documentation can be found here: https://ingest-api-dev-aueast.relevance.ai/latest/documentation#operation/InsertEncode

        - When inserting the document you can optionally specify your own id for a document by using the field name "_id", if not specified a random id is assigned.
        - When inserting or specifying vectors in a document use the suffix (ends with) "_vector_" for the field name. e.g. "product_description_vector_".
        - When inserting or specifying chunks in a document the suffix (ends with) "_chunk_" for the field name. e.g. "products_chunk_".
        - When inserting or specifying chunk vectors in a document's chunks use the suffix (ends with) "_chunkvector_" for the field name. e.g. "products_chunk_.product_description_chunkvector_".
        - Try to keep each batch of documents to insert under 200mb to avoid the insert timing out.

        Parameters
        ----------
        dataset_id : string
            Unique name of dataset
        documents : list
            A list of documents. Document is a JSON-like data that we store our metadata and vectors with. For specifying id of the document use the field '_id', for specifying vector field use the suffix of '_vector_'
        insert_date : bool
            Whether to include insert date as a field 'insert_date_'.
        overwrite : bool
            Whether to overwrite document if it exists.
        update_schema : bool
            Whether the api should check the documents for vector datatype to update the schema.
        include_inserted_ids: bool
            Include the inserted IDs in the response
        field_transformers: list
            An example field_transformers object:

            >>> {
            >>>    "field": "string",
            >>>    "output_field": "string",
            >>>    "remove_html": true,
            >>>    "split_sentences": true
            >>> }
        """
        field_transformers = [] if field_transformers is None else field_transformers

        base_url = self.config.get_option("api.base_ingest_url")

        if return_documents is False:
            return self.make_http_request(
                endpoint=f"/datasets/{dataset_id}/documents/bulk_insert",
                base_url=base_url,
                method="POST",
                parameters={
                    "documents": documents,
                    "insert_date": insert_date,
                    "overwrite": overwrite,
                    "update_schema": update_schema,
                    "field_transformers": field_transformers,
                },
            )

        else:
            response_json = self.make_http_request(
                endpoint=f"/datasets/{dataset_id}/documents/bulk_insert",
                base_url=base_url,
                method="POST",
                parameters={
                    "documents": documents,
                    "insert_date": insert_date,
                    "overwrite": overwrite,
                    "update_schema": update_schema,
                    "field_transformers": field_transformers,
                },
            )

            try:
                status_code = response_json.status_code
            except:
                status_code = 200

            return {
                "response_json": response_json,
                "documents": documents,
                "status_code": status_code,
            }

    async def bulk_insert_async(
        self,
        dataset_id: str,
        documents: list,
        insert_date: bool = True,
        overwrite: bool = True,
        update_schema: bool = True,
        field_transformers: Optional[list] = None,
    ):
        """
        Asynchronous version of bulk_insert. See bulk_insert for details.

        Parameters
        ----------
        dataset_id: str
            Unique name of dataset

        documents: list
            A list of documents. A document is a JSON-like data that we store our metadata and vectors with. For specifying id of the document use the field '_id', for specifying vector field use the suffix of '_vector_'

        insert_date: bool
            Whether to include insert date as a field 'insert_date_'.

        overwrite: bool
            Whether to overwrite document if it exists.

        update_schema: bool
            Whether the api should check the documents for vector datatype to update the schema.

        field_transformers: list
        """
        field_transformers = [] if field_transformers is None else field_transformers

        return await self.make_async_http_request(
            base_url=self.config.get_option("api.base_ingest_url"),
            endpoint=f"/datasets/{dataset_id}/documents/bulk_insert",
            method="POST",
            parameters={
                "documents": documents,
                "insert_date": insert_date,
                "overwrite": overwrite,
                "update_schema": update_schema,
                "field_transformers": field_transformers,
            },
        )

    def delete(self, dataset_id: str, confirm: bool = False):
        """
        Delete a dataset

        Parameters
        ----------
        dataset_id : string
            Unique name of dataset

        """
        if confirm:
            # confirm with the user
            self.logger.critical(f"You are about to delete {dataset_id}")
            user_input = input("Confirm? [Y/N] ")
        else:
            user_input = "y"
        # input validation
        if user_input.lower() in ("y", "yes"):
            if "gateway-api-aueast" in self.config["api.base_url"]:
                return self.make_http_request(
                    endpoint=f"/datasets/delete",
                    method="POST",
                    parameters={"dataset_id": dataset_id},
                    raise_error=False,
                )
            else:
                return self.make_http_request(
                    endpoint=f"/datasets/{dataset_id}/delete",
                    method="POST",
                    raise_error=False
                    # parameters={"dataset_id": dataset_id},
                )

        elif user_input.lower() in ("n", "no"):
            self.logger.critical(f"{dataset_id} not deleted")
            return

        else:
            self.logger.critical(f"Error: Input {user_input} unrecognised.")
            return

    def clone(
        self,
        old_dataset: str,
        new_dataset: str,
        schema: Optional[dict] = None,
        rename_fields: Optional[dict] = None,
        remove_fields: Optional[list] = None,
        filters: Optional[list] = None,
    ):
        """
        Clone a dataset into a new dataset. You can use this to rename fields and change data schemas. This is considered a project job.

        Parameters
        ----------
        old_dataset : string
            Unique name of old dataset to copy from
        new_dataset : string
            Unique name of new dataset to copy to
        schema : dict
            Schema for specifying the field that are vectors and its length
        rename_fields : dict
            Fields to rename {'old_field': 'new_field'}. Defaults to no renames
        remove_fields : list
            Fields to remove ['random_field', 'another_random_field']. Defaults to no removes
        filters : list
            Query for filtering the search results
        """
        schema = {} if schema is None else schema
        rename_fields = {} if rename_fields is None else rename_fields
        remove_fields = [] if remove_fields is None else remove_fields
        filters = [] if filters is None else filters

        dataset_id = old_dataset
        return self.make_http_request(
            endpoint=f"/datasets/{dataset_id}/clone",
            method="POST",
            parameters={
                "new_dataset_id": new_dataset,
                "schema": schema,
                "rename_fields": rename_fields,
                "remove_fields": remove_fields,
                "filters": filters,
            },
        )

    def search(
        self,
        query,
        sort_by_created_at_date: bool = False,
        asc: bool = False,
    ):
        """
        Search datasets by their names with a traditional keyword search.

        Parameters
        ----------
        query : string
            Any string that belongs to part of a dataset.
        sort_by_created_at_date : bool
            Sort by created at date. By default shows the newest datasets. Set asc=False to get oldest dataset.
        asc : bool
            Whether to sort results by ascending or descending order
        """
        return self.make_http_request(
            endpoint="/datasets/search",
            method="GET",
            parameters={
                "query": query,
                "sort_by_created_at_date": sort_by_created_at_date,
                "asc": asc,
            },
        )

    def task_status(self, dataset_id: str, task_id: str):
        """
        Check the status of an existing encoding task on the given dataset. \n

        The required task_id was returned in the original encoding request such as datasets.vectorize.

        Parameters
        ----------
        dataset_id : string
            Unique name of dataset
        task_id : string
            The task ID of the earlier queued vectorize task

        """
        return self.make_http_request(
            endpoint=f"/datasets/{dataset_id}/task_status",
            method="GET",
            parameters={"task_id": task_id},
        )

    def get_file_upload_urls(self, dataset_id: str, files: List):
        """
        Specify a list of file paths. For each file path, a url upload_url is returned. files can be POSTed on upload_url to upload them. They can then be accessed on url. Upon dataset deletion, these files will be deleted.

        Parameters
        -------------
        files: list
            List of files to be uploaded
        dataset_id: str
            The dataset
        """
        return self.make_http_request(
            endpoint=f"/datasets/{dataset_id}/get_file_upload_urls",
            method="POST",
            parameters={"files": files},
        )



import random

from datetime import datetime

import numpy as np
import pandas as pd

from relevanceai.utils import make_id


def complex_nested_document():
    document = {
        "sample_1": {
            "panda": pd.DataFrame(
                np.random.randint(0, 20, size=(20, 4)), columns=list("ABCD")
            ),
            "datetime": datetime.now(),
            "numpy": np.random.rand(3, 2),
            "test1": random.random(),
            "test2": random.random(),
        },
        "sample_2": {
            "subsample1": {
                "panda": pd.DataFrame(
                    np.random.randint(0, 20, size=(20, 4)), columns=list("ABCD")
                ),
                "datetime": datetime.now(),
                "numpy": np.random.rand(3, 2),
            },
            "subsample2": {
                "panda": pd.DataFrame(
                    np.array([[1, 2, np.nan], [4, np.nan, 6], [7, 8, 9]]),
                    columns=["a", "b", "c"],
                ),
                "datetime": datetime.now(),
                "numpy": np.random.rand(3, 2),
            },
        },
    }
    document["_id"] = make_id(document)
    return document


def simple_nested_document():
    document = {
        "col1": {"subcol1": random.random(), "subcol2": random.random()},
        "col2": {"subcol3": random.random(), "subcol4": random.random()},
        "col3": {"subcol5": random.random(), "subcol6": random.random()},
        "col4": {"subcol7": random.random(), "subcol8": random.random()},
    }
    document["_id"] = make_id(document)
    return document



class Messages:

    BETA_DOCSTRING = """{}

    .. warning::
        This function is currently in beta and is liable to change in the future. We
        recommend not using this in production systems.

    """
    BETA = "This function currently in beta and may change in the future."

    ADDED_DOCSTRING = """

    .. note::
        This function was introduced in **{}**.

    """

    DEPRECEATED_DOCSTRING = """

    .. note::
        This function has been deprecated as of {}

    """
    DEPRECEATED = "Deprecated. Revert to versions before {} for function. {}"

    FAIL_MESSAGE = """Your API key is invalid. Please login again"""

    WELCOME_MESSAGE = """Welcome to RelevanceAI. Logged in as {}."""

    TOKEN_MESSAGE = "Activation token (you can find it here: {} )\n"



import os
import random
import string

from typing import List

TEST_TOKEN = os.getenv("TEST_TOKEN")
if TEST_TOKEN is None:
    PROJECT = os.getenv("TEST_PROJECT")
    API_KEY = os.getenv("TEST_API_KEY")

    if PROJECT is None and API_KEY is None:
        raise ValueError("Set Env Var TEST_TOKEN")
    else:
        TEST_TOKEN = f"{PROJECT}:{API_KEY}:us-east-1:{None}"

TEST_TOKEN = TEST_TOKEN.split(":")
TEST_FIREBASE_UID = "relevanceai-sdk-test-user"
TEST_TOKEN[-1] = TEST_FIREBASE_UID
TEST_TOKEN = ":".join(TEST_TOKEN)


VECTOR_LENGTH = 8
NUMBER_OF_DOCUMENTS = int(os.getenv("TEST_NUMBER_OF_DOCUMENTS", 20))


def generate_random_string(string_length: int = 5) -> str:
    """Generate a random string of letters and numbers"""
    return "".join(
        random.choice(string.ascii_uppercase + string.digits)
        for _ in range(string_length)
    )


def generate_random_vector(vector_length: int = VECTOR_LENGTH) -> List[float]:
    """Generate a random list of floats"""
    return [random.random() for _ in range(vector_length)]


def generate_random_label(label_value: int = 5) -> str:
    return f"label_{random.randint(0, label_value)}"


def generate_random_integer(min: int = 0, max: int = 100) -> int:
    return random.randint(min, max)


def generate_dataset_id():
    return SAMPLE_DATASET_DATASET_PREFIX + generate_random_string().lower()


RANDOM_PANDAS_DATASET_SUFFIX = generate_random_string().lower()
SAMPLE_DATASET_DATASET_PREFIX = "_sample_test_dataset_"
LABEL_DATSET_ID = SAMPLE_DATASET_DATASET_PREFIX + generate_random_string().lower()



import requests
import numpy as np
import pandas as pd

from sklearn.cluster import KMeans

from relevanceai.reports.cluster.report import ClusterReport


def test_cluster_reporting_smoke():
    docs = requests.get(
        "https://raw.githubusercontent.com/fanzeyi/pokemon.json/master/pokedex.json"
    ).json()
    for d in docs:
        b = d["base"]
        d.update(b)
        d["base_vector_"] = [
            b["Attack"],
            b["Defense"],
            b["HP"],
            b["Sp. Attack"],
            b["Sp. Defense"],
            b["Speed"],
        ]

    df = pd.DataFrame(docs)
    vectors = np.array(df["base_vector_"].tolist())

    n_clusters = 2
    kmeans = KMeans(n_clusters=n_clusters)
    cluster_labels = kmeans.fit_predict(vectors)

    report = ClusterReport(
        X=vectors, cluster_labels=cluster_labels, num_clusters=n_clusters, model=kmeans
    )

    internal_report = report.internal_report
    assert len(internal_report) >= 0



import pytest

from sklearn.cluster import KMeans

from relevanceai import ClusterOps

from relevanceai.dataset import Dataset

from .utils import VECTOR_FIELDS


@pytest.fixture(scope="function")
def kmeans_clusterer(test_dataset: Dataset):
    clusterer: ClusterOps = test_dataset.cluster(
        model=KMeans(n_clusters=2),
        alias="kmeans-2",
        vector_fields=VECTOR_FIELDS,
    )
    yield clusterer
    clusterer.delete_centroids(test_dataset.dataset_id, VECTOR_FIELDS)



"""Report for grading
"""


def get_silhouette_grade(silhouette_score):
    if silhouette_score > 0.9:
        return "S"
    if silhouette_score > 0.7:
        return "A"
    if silhouette_score > 0.6:
        return "B"
    if silhouette_score > 0.4:
        return "C"
    if silhouette_score > 0.2:
        return "D"
    if silhouette_score > 0.0:
        return "D"
    if silhouette_score > -0.2:
        return "E"
    if silhouette_score > -0.4:
        return "E"
    if silhouette_score > -0.6:
        return "F"
    if silhouette_score > -0.8:
        return "F"
    return "F"


def dunn_index(X, y):
    raise NotImplementedError("Dunn index not supported at the moment.")


# def get_dunn_index(score):
#     _internal_report = {}
#     _internal_report["overall"]["dunn_index"] = self.dunn_index(
#         min_centroid_distance, max_centroid_distance
#     )
#     pass


def get_dunn_index_grade(score):
    if score > 0.9:
        return "AA"
    if score > 0.8:
        return "A"
    if score > 0.7:
        return "B"
    if score > 0.6:
        return "C"
    if score > 0.5:
        return "D"
    if score > 0.4:
        return "E"
    return "F"



from relevanceai.dataset import Dataset


def test_series_sample(test_dataset: Dataset):
    sample_n = test_dataset["sample_1_label"].sample(n=10)
    assert len(sample_n) == 10


def test_series_sample(test_dataset: Dataset):
    sample_n = test_dataset[["sample_1_label", "sample_2_label"]].sample(n=10)
    assert len(sample_n) == 10
    assert len(sample_n[0].keys()) == 3


def test_series_add(test_dataset: Dataset):
    series_1 = test_dataset["sample_1_value"]
    series_2 = test_dataset["sample_2_value"]
    series_3 = test_dataset["sample_1_label"]

    series_1 + series_2

    try:
        series_1 + series_3
    except ValueError:
        assert True

    try:
        series_3 + series_2
    except ValueError:
        assert True



# TODO: Update this into a workflow
# import numpy as np
# import warnings
# from random import randint
# from typing import List, Tuple
# from sklearn import preprocessing
# from tqdm.auto import tqdm
# from typing import List, Optional, Union
# from doc_utils import DocUtils
# from relevanceai.package_utils.logger import LoguruLogger
# from relevanceai.api.client import BatchAPIClient
# from relevanceai.unstructured_data.text.base_text_processing import BaseTextProcessing


# class PlotTextThemeModel(BatchAPIClient, BaseTextProcessing, LoguruLogger, DocUtils):
#     def __init__(
#         self,
#         project: str,
#         api_key: str,
#         firebase_uid: str,
#         # dataset_info
#         dataset_id: str,
#         upload_chunksize: int = 50,
#         # clustering
#         cluster_field: str = "_cluster_",
#         # dimensionality reduction parameters
#         embedding_dims: int = 2,
#         dim_red_k: int = 800,
#         n_epochs_without_progress: int = 100,
#         # nlp parameters
#         language: str = "english",
#     ):
#         self.dataset_id = dataset_id
#         self.upload_chunksize = upload_chunksize
#         self.cluster_field = cluster_field
#         self.embedding_dims = embedding_dims
#         self.dim_red_k = dim_red_k
#         self.n_epochs_without_progress = n_epochs_without_progress
#         self.language = language
#         super().__init__(project=project, api_key=api_key, firebase_uid=firebase_uid)

#     def _build_and_plot_clusters(
#         self,
#         vector_fields: List[str],
#         text_fields: List[str],
#         max_doc_num: int = None,
#         k: int = 10,
#         alias: str = "kmeans",
#         lower: bool = True,
#         remove_digit: bool = True,
#         remove_punct: bool = True,
#         remove_stop_words: bool = True,
#         additional_stop_words: Optional[List[str]] = None,
#         cluster_representative_cnt: int = 3,
#         plot_axis: str = "off",
#         figsize: Tuple[int, ...] = (20, 10),
#         cmap: str = "plasma",
#         alpha: float = 0.2,
#     ):
#         additional_stop_words = (
#             [] if additional_stop_words is None else additional_stop_words
#         )

#         # get documents
#         documents = self._get_documents(
#             vector_fields=vector_fields,
#             text_fields=text_fields,
#             max_doc_num=max_doc_num,
#         )

#         # perform kmeans clustering
#         alias = alias + "_" + str(k)
#         centers, clustered_documents = self._kmeans_clustering(
#             documents=documents, vector_fields=vector_fields, k=k, alias=alias
#         )

#         clustering_results = self.get_field_across_documents(
#             ".".join([self.cluster_field, ".".join(vector_fields), alias]), documents
#         )
#         for i, cl_doc in enumerate(tqdm(clustered_documents)):
#             documents[i][
#                 "_".join([self.cluster_field, "".join(vector_fields), alias])
#             ] = clustering_results[i]

#         # get cluster data
#         cluster_data = self._get_cluster_datafield(
#             documents=documents,
#             vector_fields=vector_fields,
#             text_fields=text_fields,
#             alias=alias,
#             lower=lower,
#             remove_digit=remove_digit,
#             remove_punct=remove_punct,
#         )
#         clusters_top_n_words = self._get_cluster_word_freq(
#             cluster_data=cluster_data,
#             remove_stop_words=remove_stop_words,
#             additional_stop_words=additional_stop_words,
#             cluster_representative_cnt=cluster_representative_cnt,
#         )
#         for c in clusters_top_n_words:
#             self.logger.info(c, clusters_top_n_words[c])

#         # dimensionality reduction
#         # 1) Fit the model and reduce vector size
#         vectors: List = [[] for _ in documents]
#         for vector_field in vector_fields:
#             t = self.get_field_across_documents(vector_field, documents)
#             vectors = [i + j for i, j in zip(vectors, t)]  # concatenate vectors
#         dr_documents = self._dim_reduction(
#             vector_data=vectors,
#             embedding_dims=self.embedding_dims,
#             k=self.dim_red_k,
#             n_epochs_without_progress=self.n_epochs_without_progress,
#         )
#         for i, dr in enumerate(dr_documents):
#             documents[i][vector_field + "_dr_vector_"] = dr.tolist()
#         # 2) reduce center vector size
#         vectors = [c["centroid_vector_"] for c in centers]
#         dr_centers = self._dim_reduction(
#             vector_data=vectors,
#             embedding_dims=self.embedding_dims,
#             k=self.dim_red_k,
#             n_epochs_without_progress=self.n_epochs_without_progress,
#             run_fit=False,
#         )
#         for i, c in enumerate(centers):
#             c["centroid_dr_vector_"] = dr_centers[i]

#         # plot
#         self._plot_clusters(
#             documents,
#             dr_documents,
#             centers,
#             cluster_data,
#             vector_fields,
#             alias,
#             cluster_representative_cnt,
#             plot_axis=plot_axis,
#             figsize=figsize,
#             cmap=cmap,
#             alpha=alpha,
#         )

#     def _get_documents(
#         self, vector_fields: List[str], text_fields: List[str], max_doc_num: int = None
#     ):
#         self.logger.info(" * Loading documents")
#         filters = [
#             {
#                 "field": vector_field,
#                 "filter_type": "exists",
#                 "condition": "==",
#                 "condition_value": " ",
#             }
#             for vector_field in vector_fields
#         ]
#         fields = vector_fields + text_fields
#         if max_doc_num:
#             page_size = 200 if max_doc_num > 200 else max_doc_num
#             batch_doc = self._batch_load_documents(
#                 fields=fields, filters=filters, page_size=page_size
#             )
#             documents = batch_doc["documents"]
#             cursor = batch_doc["cursor"]
#             while (
#                 batch_doc["documents"] != []
#                 and len(documents) > max_doc_num - page_size
#             ):
#                 batch_doc = self._batch_load_documents(
#                     fields=fields, filters=filters, page_size=page_size, cursor=cursor
#                 )
#                 documents.extend(batch_doc["documents"])
#         else:
#             documents = self.get_all_documents(
#                 dataset_id=self.dataset_id, filters=filters, select_fields=fields
#             )
#         return documents

#     def _batch_load_documents(
#         self,
#         fields: List[str],
#         filters: Optional[List[dict]] = None,
#         page_size: int = 200,
#         cursor: str = None,
#     ):
#         return self.datasets.documents.get_where(
#             dataset_id=self.dataset_id,
#             select_fields=fields,
#             page_size=page_size,
#             filters=[] if filters is None else filters,
#             cursor=cursor,
#         )

#     def _kmeans_clustering(
#         self,
#         documents: List[dict],
#         vector_fields: List[str],
#         k: int = 10,
#         alias: str = "kmeans",
#     ):
#         self.logger.info("*Kmeans Clustering")
#         clusterer = self.ClusterOps()
#         clustered_documents = clusterer.fit_documents(
#             vector_fields, documents, alias=alias
#         )
#         res = self._update_documents(
#             self.dataset_id, clustered_documents, chunksize=self.upload_chunksize
#         )
#         centers = clusterer.get_centroid_documents()
#         self.services.cluster.centroids.insert(
#             dataset_id=self.dataset_id,
#             cluster_centers=centers,
#             vector_fields=vector_fields,
#             alias=alias,
#         )
#         return centers, clustered_documents

#     @staticmethod
#     def _get_fields_value_by_id(documents: List[dict], id: str, fields: List[str]):
#         vals = {}
#         for doc in documents:
#             if doc["_id"] == id:
#                 for field in fields:
#                     vals[field] = DocUtils.get_field(field, doc) + " "
#         return vals

#     def _get_cluster_datafield(
#         self,
#         documents: List[dict],
#         vector_fields: List[str],
#         text_fields: List[str],
#         alias: str,
#         lower: bool = True,
#         remove_digit: bool = True,
#         remove_punct: bool = True,
#     ):
#         self.logger.info(" * Extracting cluster info")
#         cluster_data: dict = {}
#         for i, doc in tqdm(enumerate(documents)):
#             cluster_name = doc[
#                 "_".join([self.cluster_field, "".join(vector_fields), alias])
#             ]
#             if cluster_name not in cluster_data:
#                 cluster_data[cluster_name] = {"data": []}
#             values = self._get_fields_value_by_id(
#                 documents, id=doc["_id"], fields=text_fields
#             )
#             text = self.normalize_text(
#                 txt=" ".join([v for k, v in values.items()]),
#                 lower=lower,
#                 remove_digit=remove_digit,
#                 remove_punct=remove_punct,
#             )
#             cluster_data[cluster_name]["data"].append(text)
#         return cluster_data

#     @staticmethod
#     def _get_cluster_population(cluster_data: dict):
#         return [
#             (cluster_name, len(v["data"])) for cluster_name, v in cluster_data.items()
#         ]

#     def _get_cluster_word_freq(
#         self,
#         cluster_data: dict,
#         remove_stop_words: bool = True,
#         additional_stop_words: List[str] = [],
#         cluster_representative_cnt: int = 3,
#     ):
#         for cluster in tqdm(cluster_data):
#             cluster_data[cluster]["word_freq"] = self.get_word_frequency(
#                 str_list=cluster_data[cluster]["data"],
#                 remove_stop_words=remove_stop_words,
#                 additional_stop_words=[]
#                 if additional_stop_words is None
#                 else additional_stop_words,
#                 language=self.language,
#             )
#         return {
#             c: [cluster_data[c]["word_freq"][:cluster_representative_cnt]]
#             for c in cluster_data
#         }

#     def _dim_reduction(
#         self,
#         vector_data: Union[List, np.ndarray],
#         embedding_dims=2,
#         k=800,
#         n_epochs_without_progress=100,
#         run_fit=True,
#     ):

#         try:
#             from ivis import Ivis
#         except ModuleNotFoundError:
#             warnings.warn(
#                 """You are missing ivis, please run the appropriate installation option:
#             `pip install ivis[gpu]` If you have CUDA installed
#             `pip install ivis[cpu]` if you don't have CUDA installed
#             """
#             )

#         self.logger.info(" * Dimensionality reduction")
#         vector_data = np.array(vector_data)
#         if run_fit:
#             self.dr_model = Ivis(
#                 embedding_dims=embedding_dims,
#                 k=k,
#                 n_epochs_without_progress=n_epochs_without_progress,
#             )
#             dr_documents = self.dr_model.fit_transform(vector_data)
#         else:
#             dr_documents = self.dr_model.transform(vector_data)

#         return dr_documents

#     def _plot_clusters(
#         self,
#         documents: List[dict],
#         dr_documents: List,
#         centers: List[dict],
#         cluster_data: dict,
#         vector_fields: List[str],
#         alias: str,
#         cluster_representative_cnt: int,
#         plot_axis: str = "off",
#         figsize: Tuple[int, ...] = (20, 10),
#         cmap: str = "plasma",
#         alpha: float = 0.2,
#     ):
#         group = [
#             x["_".join([self.cluster_field, "".join(vector_fields), alias])]
#             for x in documents
#         ]
#         le = preprocessing.LabelEncoder()
#         color = le.fit_transform(group)

#         try:
#             import matplotlib.pyplot as plt
#         except ModuleNotFoundError:
#             warnings.warn(
#                 "You are missing matplotlib, please run `pip install matplotlib`"
#             )

#         plt.axis(plot_axis)
#         plt.figure(figsize=figsize)
#         plt.scatter(
#             [x[0] for x in dr_documents],
#             [x[1] for x in dr_documents],
#             c=color,
#             cmap=cmap,
#             alpha=alpha,
#         )

#         for i, r in enumerate(centers):
#             plt.annotate(
#                 [
#                     x[0]
#                     for x in cluster_data[r["_id"]]["word_freq"][
#                         :cluster_representative_cnt
#                     ]
#                 ],
#                 (r["centroid_dr_vector_"][0] - 0.5, r["centroid_dr_vector_"][1]),
#             )


# def build_and_plot_clusters(
#     self,
#     project: str,
#     api_key: str,
#     firebase_uid: str,
#     dataset_id: str,
#     vector_fields: List[str],
#     text_fields: List[str],
#     upload_chunksize: int = 50,
#     cluster_field: str = "_cluster_",
#     embedding_dims: int = 2,
#     dim_red_k: int = 800,
#     n_epochs_without_progress: int = 100,
#     language: str = "english",
#     max_doc_num: int = None,
#     k: int = 10,
#     alias: str = "kmeans",
#     lower: bool = True,
#     remove_digit: bool = True,
#     remove_punct: bool = True,
#     remove_stop_words: bool = True,
#     additional_stop_words: Optional[List[str]] = None,
#     cluster_representative_cnt: int = 3,
#     plot_axis: str = "off",
#     figsize: Tuple[int, ...] = (20, 10),
#     cmap: str = "plasma",
#     alpha: float = 0.2,
# ):
#     model = PlotTextThemeModel(
#         project=project,
#         api_key=api_key,
#         firebase_uid=firebase_uid,
#         dataset_id=dataset_id,
#         upload_chunksize=upload_chunksize,
#         # clustering
#         cluster_field=cluster_field,
#         # dimensionality reduction parameters
#         embedding_dims=embedding_dims,
#         dim_red_k=dim_red_k,
#         n_epochs_without_progress=n_epochs_without_progress,
#         language=language,
#     )
#     return model._build_and_plot_clusters(
#         vector_fields=vector_fields,
#         text_fields=text_fields,
#         max_doc_num=max_doc_num,
#         k=k,
#         alias=alias,
#         lower=lower,
#         remove_digit=remove_digit,
#         remove_punct=remove_punct,
#         remove_stop_words=remove_stop_words,
#         additional_stop_words=[]
#         if additional_stop_words is None
#         else additional_stop_words,
#         cluster_representative_cnt=cluster_representative_cnt,
#         plot_axis=plot_axis,
#         figsize=figsize,
#         cmap=cmap,
#         alpha=alpha,
#     )



import pytest

from relevanceai.utils.integration_checks import is_sklearn_available

if is_sklearn_available():
    from sklearn.cluster import KMeans

from typing import Dict, List, NamedTuple

from relevanceai import Client

from tests.globals.constants import generate_dataset_id


@pytest.fixture(scope="function")
def obj_dataset_id(test_client: Client, dataclass_documents: List[NamedTuple]):
    test_dataset_id = generate_dataset_id()

    test_client._insert_documents(test_dataset_id, dataclass_documents)

    yield test_dataset_id

    test_client.datasets.delete(test_dataset_id)


@pytest.fixture(scope="function")
def clustered_dataset_id(test_client: Client, vector_documents: List[NamedTuple]):
    test_dataset_id = generate_dataset_id()

    test_client._insert_documents(test_dataset_id, vector_documents)
    dataset = test_client.Dataset(test_dataset_id)
    dataset.cluster(
        model=KMeans(n_clusters=10),
        vector_fields=["sample_1_vector_"],
    )
    yield test_dataset_id

    test_client.datasets.delete(test_dataset_id)


@pytest.fixture(scope="function")
def test_clusterer(test_client: Client, vector_documents: List[Dict]):
    test_dataset_id = generate_dataset_id()

    test_client._insert_documents(test_dataset_id, vector_documents)

    dataset = test_client.Dataset(test_dataset_id)
    clusterer = dataset.cluster(
        model=KMeans(n_clusters=10),
        vector_fields=["sample_1_vector_"],
        alias="kmeans-10",
    )
    yield clusterer

    test_client.datasets.delete(test_dataset_id)



import pytest

from typing import List, Dict

from relevanceai import Client

from tests.globals.constants import generate_dataset_id


@pytest.fixture(scope="function")
def assorted_nested_dataset_id(
    test_client: Client, assorted_nested_documents: List[Dict]
):
    test_dataset_id = generate_dataset_id()

    test_client._insert_documents(test_dataset_id, assorted_nested_documents)

    yield test_dataset_id

    test_client.datasets.delete(test_dataset_id)



"""
All read operations for Dataset
"""
import re
import math
import warnings
import pandas as pd

from typing import Optional, Union, Dict, List
from relevanceai.client.helpers import Credentials

from relevanceai.operations.cluster.centroids import Centroids

from relevanceai.dataset.read.metadata import Metadata
from relevanceai.dataset.read.statistics import Statistics
from relevanceai.dataset.helpers import _build_filters

from relevanceai.utils.cache import lru_cache
from relevanceai.utils.decorators.analytics import track

from relevanceai.constants.constants import MAX_CACHESIZE
from relevanceai.constants.warning import Warning


class Read(Statistics):
    """

    Dataset Read
    -------------------

    A Pandas Like datatset API for interacting with the RelevanceAI python package
    """

    def __init__(
        self,
        credentials: Credentials,
        dataset_id: str,
        fields: Optional[list] = None,
        image_fields: Optional[List[str]] = None,
        audio_fields: Optional[List[str]] = None,
        text_fields: Optional[List[str]] = None,
        highlight_fields: Optional[Dict[str, list]] = None,
        **kwargs,
    ):
        self.credentials = credentials
        self.fields = [] if fields is None else fields
        self.dataset_id = dataset_id
        self.centroids = Centroids(
            credentials=credentials,
            dataset_id=self.dataset_id,
        )
        self.image_fields = [] if image_fields is None else image_fields
        self.audio_fields = [] if audio_fields is None else audio_fields
        self.text_fields = [] if text_fields is None else text_fields
        self.highlight_fields = {} if highlight_fields is None else highlight_fields
        super().__init__(
            credentials=credentials,
            dataset_id=dataset_id,
            **kwargs,
        )

    @property  # type: ignore
    @track
    def shape(self):
        """
        Returns the shape (N x C) of a dataset
        N = number of samples in the Dataset
        C = number of columns in the Dataset

        Returns
        -------
        Tuple
            (N, C)

        Example
        ---------------
        .. code-block::

            from relevanceai import Client

            client = Client()

            dataset_id = "sample_dataset_id"
            df = client.Dataset(dataset_id)

            length, width = df.shape
        """
        schema = self.datasets.schema(self.dataset_id)
        n_documents = self.get_number_of_documents(dataset_id=self.dataset_id)
        return (n_documents, len(schema))

    def _get_possible_dtypes(self, schema):
        possible_dtypes = []
        for v in schema.values():
            if isinstance(v, str):
                possible_dtypes.append(v)
            elif isinstance(v, dict):
                if list(v)[0] == "vector":
                    possible_dtypes.append("vector_")
        return possible_dtypes

    def _get_dtype_count(self, schema: dict):
        possible_dtypes = self._get_possible_dtypes(schema)
        dtypes = {
            dtype: list(schema.values()).count(dtype) for dtype in possible_dtypes
        }
        return dtypes

    def _get_schema(self):
        # stores schema in memory to save users API usage/reloading
        if hasattr(self, "_schema"):
            return self._schema
        self._schema = self.datasets.schema(self.dataset_id)
        return self._schema

    @track
    def info(self, dtype_count: bool = False) -> pd.DataFrame:
        """
        Return a dictionary that contains information about the Dataset
        including the index dtype and columns and non-null values.

        Parameters
        -----------
        dtype_count: bool
            If dtype_count is True, prints a value_counts of the data type


        Returns
        ---------
        pd.DataFrame
            a pandas dataframe of information

        Example
        ---------------
        .. code-block::

            from relevanceai import Client

            client = Client()

            dataset_id = "sample_dataset_id"
            df = client.Dataset(dataset_id)
            df.info()
        """
        health: dict = self.datasets.monitor.health(self.dataset_id)
        schema: dict = self._get_schema()
        info_json = [
            {
                "Column": column,
                "Dtype": schema[column],
            }
            if column not in health
            else {
                "Column": column,
                "Non-Null Count": health[column]["missing"],
                "Dtype": schema[column],
            }
            for column in schema
        ]

        info_df = pd.DataFrame(info_json)
        if dtype_count:
            dtypes_info = self._get_dtype_count(schema)
            print(dtypes_info)
        return info_df

    @track
    def head(
        self, n: int = 5, raw_json: bool = False, **kw
    ) -> Union[dict, pd.DataFrame]:
        """
        Return the first `n` rows.
        returns the first `n` rows of your dataset.
        It is useful for quickly testing if your object
        has the right type of data in it.

        Parameters
        ----------
        n : int, default 5
            Number of rows to select.
        raw_json: bool
            If True, returns raw JSON and not Pandas Dataframe
        kw:
            Additional arguments to feed into show_json

        Returns
        -------
        Pandas DataFrame or Dict, depending on args
            The first 'n' rows of the caller object.

        Example
        ---------
        .. code-block::

            from relevanceai import Client

            client = Client()

            df = client.Dataset("sample_dataset_id", image_fields=["image_url])

            df.head()
        """
        print(
            f"https://cloud.relevance.ai/dataset/{self.dataset_id}/dashboard/data?page=1"
        )
        head_documents = self.get_documents(
            number_of_documents=n,
        )
        if raw_json:
            return head_documents
        else:
            try:
                return self._show_json(head_documents, **kw)
            except Exception as e:
                warnings.warn(Warning.MISSING_RELEVANCE_NOTEBOOK + str(e))
                return pd.json_normalize(head_documents).head(n=n)

    def _show_json(self, documents, **kw):
        from jsonshower import show_json

        if not self.text_fields:
            text_fields = pd.json_normalize(documents).columns.tolist()
        else:
            text_fields = self.text_fields
        return show_json(
            documents,
            image_fields=self.image_fields,
            audio_fields=self.audio_fields,
            highlight_fields=self.highlight_fields,
            text_fields=text_fields,
            **kw,
        )

    def _repr_html_(self):
        documents = self.get_documents()
        documents = [
            {
                "_id": document["_id"],
                "insert_date_": document["insert_date_"],
                **document,
            }
            for document in documents
        ]
        try:
            return self._show_json(documents, return_html=True)
        except Exception as e:
            warnings.warn(Warning.MISSING_RELEVANCE_NOTEBOOK + str(e))
            return pd.json_normalize(documents).set_index("_id")._repr_html_()

    @track
    def sample(
        self,
        n: int = 1,
        frac: float = None,
        filters: Optional[list] = None,
        random_state: int = 0,
        select_fields: Optional[list] = None,
        include_vector: bool = True,
        output_format: str = "json",
    ):
        """
        Return a random sample of items from a dataset.

        Parameters
        ----------
        n : int
            Number of items to return. Cannot be used with frac.
        frac: float
            Fraction of items to return. Cannot be used with n.
        filters: list
            Query for filtering the search results
        random_state: int
            Random Seed for retrieving random documents.
        select_fields: list
            Fields to include in the search results, empty array/list means all fields.

        Example
        ---------

        .. code-block::

            from relevanceai import Client
            client = Client()
            df = client.Dataset("sample_dataset_id", image_fields=["image_url])
            df.sample()
        """
        filters = [] if filters is None else filters
        select_fields = [] if select_fields is None else select_fields

        if not select_fields and self.fields:
            select_fields = self.fields

        if frac and n:
            raise ValueError("Only one of n or frac can be provided")

        if frac:
            if frac > 1 or frac < 0:
                raise ValueError("Fraction must be between 0 and 1")
            n = math.ceil(
                self.get_number_of_documents(self.dataset_id, filters=filters) * frac
            )

        documents = self.datasets.documents.get_where(
            dataset_id=self.dataset_id,
            filters=filters,
            page_size=n,
            random_state=random_state,
            is_random=True,
            select_fields=select_fields,
            include_vector=include_vector,
        )["documents"]
        if output_format == "json":
            return documents
        elif output_format == "pandas":
            return pd.DataFrame.from_dict(documents, orient="records")

    @lru_cache(maxsize=MAX_CACHESIZE)
    @track
    def get_all_documents(
        self,
        chunksize: int = 1000,
        filters: Optional[List] = None,
        sort: Optional[List] = None,
        select_fields: Optional[List] = None,
        include_vector: bool = True,
        show_progress_bar: bool = True,
    ):
        """
        Retrieve all documents with filters. Filter is used to retrieve documents that match the conditions set in a filter query. This is used in advance search to filter the documents that are searched. For more details see documents.get_where.

        Parameters
        ------------
        chunksize: list
            Number of documents to retrieve per retrieval
        include_vector: bool
            Include vectors in the search results
        sort: list
            Fields to sort by. For each field, sort by descending or ascending. If you are using descending by datetime, it will get the most recent ones.
        filters: list
            Query for filtering the search results
        select_fields : list
            Fields to include in the search results, empty array/list means all fields.

        Example
        ----------

        .. code-block::

            from relevanceai import Client
            client = Client()
            dataset_id = "sample_dataset_id"
            df = client.Dataset(dataset_id)
            documents = df.get_all_documents()

        """
        filters = [] if filters is None else filters
        sort = [] if sort is None else sort
        select_fields = [] if select_fields is None else select_fields

        return self._get_all_documents(
            dataset_id=self.dataset_id,
            chunksize=chunksize,
            filters=filters,
            sort=sort,
            select_fields=select_fields,
            include_vector=include_vector,
            show_progress_bar=show_progress_bar,
        )

    @track
    def get_documents_by_ids(
        self, document_ids: Union[List, str], include_vector: bool = True
    ):
        """
        Retrieve a document by its ID ("_id" field). This will retrieve the document faster than a filter applied on the "_id" field.

        Parameters
        ----------
        document_ids: Union[list, str]
            ID of a document in a dataset.
        include_vector: bool
            Include vectors in the search results

        Example
        --------

        .. code-block::

            from relevanceai import Client, Dataset
            client = Client()
            dataset_id = "sample_dataset_id"
            df = client.Dataset(dataset_id)
            df.get_documents_by_ids(["sample_id"], include_vector=False)
        """
        if isinstance(document_ids, str):
            return self.datasets.documents.get(
                self.dataset_id, id=document_ids, include_vector=include_vector
            )
        elif isinstance(document_ids, list):
            return self.datasets.documents.bulk_get(
                self.dataset_id, ids=document_ids, include_vector=include_vector
            )
        raise TypeError("Document IDs needs to be a string or a list")

    @track
    def get(self, document_ids: Union[List, str], include_vector: bool = True):
        """
        Retrieve a document by its ID ("_id" field). This will retrieve the document faster than a filter applied on the "_id" field.
        This has the same functionality as get_document_by_ids.

        Parameters
        ----------
        document_ids: Union[list, str]
            ID of a document in a dataset.
        include_vector: bool
            Include vectors in the search results

        Example
        --------

        .. code-block::

            from relevanceai import Client
            client = Client()
            dataset_id = "sample_dataset_id"
            df = client.Dataset(dataset_id)
            df.get(["sample_id"], include_vector=False)
        """
        if isinstance(document_ids, str):
            return self.datasets.documents.get(
                self.dataset_id, id=document_ids, include_vector=include_vector
            )
        elif isinstance(document_ids, list):
            return self.datasets.documents.bulk_get(
                self.dataset_id, ids=document_ids, include_vector=include_vector
            )
        raise TypeError("Document IDs needs to be a string or a list")

    @property  # type: ignore
    @track
    def schema(self) -> Dict:
        """
        Returns the schema of a dataset. Refer to datasets.create for different field types available in a VecDB schema.

        Example
        -----------------

        .. code-block::

            from relevanceai import Client
            client = Client()
            dataset_id = "sample_dataset_id"
            df = client.Dataset(dataset_id)
            df.schema
        """
        return self.datasets.schema(self.dataset_id)

    @property  # type: ignore
    @track
    def columns(self) -> List[str]:
        """
        Returns a list of columns

        Example
        ----------

        .. code-block::

            from relevanceai import Client
            client = Client()
            dataset_id = "sample_dataset_id"
            df = client.Dataset(dataset_id)
            df.columns

        """
        return list(self.schema)

    @track
    def filter(
        self,
        index: Union[str, None] = None,
        items: Union[List, None] = None,
        like: Union[str, None] = None,
        regex: Union[str, None] = None,
        axis: Union[int, str] = 0,
    ):
        """
        Returns a subset of the dataset, filtered by the parameters given

        Parameters
        ----------
        items : str, default None
            the column on which to filter, if None then defaults to the _id column
        items : list-like
            Keep labels from axis which are in items.
        like : str
            Keep labels from axis for which "like in label == True".
        regex : str (regular expression)
            Keep labels from axis for which re.search(regex, label) == True.
        axis : {0 or `index`, 1 or `columns`},
            The axis on which to perform the search

        Returns
        ---------
        list of documents

        Example
        ----------

        .. code-block::

            from relevanceai import Client
            client = Client()
            df = client.Dataset("ecommerce-example-encoded")
            filtered = df.filter(items=["product_title", "query", "product_price"])
            filtered = df.filter(index="query", like="routers")
            filtered = df.filter(index="product_title", regex=".*Hard.*Drive.*")

        """
        fields = []
        filters = []

        schema = list(self.schema)

        if index:
            axis = 0
        else:
            axis = 1
            index = "_id"

        rows = axis in [0, "index"]
        columns = axis in [1, "columns"]

        if items is not None:
            if columns:
                fields += items

            elif rows:
                filters += _build_filters(items, filter_type="exact_match", index=index)

        elif like:
            if columns:
                fields += [column for column in schema if like in column]

            elif rows:
                filters += _build_filters(like, filter_type="contains", index=index)

        elif regex:
            if columns:
                query = re.compile(regex)
                re_fields = list(filter(query.match, schema))
                fields += re_fields

            elif rows:
                filters += _build_filters(regex, filter_type="regexp", index=index)

        else:
            raise TypeError("Must pass either `items`, `like` or `regex`")

        filters = [{"filter_type": "or", "condition_value": filters}]

        return self.get_all_documents(select_fields=fields, filters=filters)

    @track
    def get_documents(
        self,
        number_of_documents: int = 20,
        filters: Optional[list] = None,
        cursor: str = None,
        batch_size: int = 1000,
        sort: Optional[list] = None,
        select_fields: Optional[list] = None,
        include_vector: bool = True,
        include_cursor: bool = False,
    ):
        """
        Retrieve documents with filters. Filter is used to retrieve documents that match the conditions set in a filter query. This is used in advance search to filter the documents that are searched. \n
        If you are looking to combine your filters with multiple ORs, simply add the following inside the query {"strict":"must_or"}.
        Parameters
        ----------
        dataset_id: string
            Unique name of dataset
        number_of_documents: int
            Number of documents to retrieve
        select_fields: list
            Fields to include in the search results, empty array/list means all fields.
        cursor: string
            Cursor to paginate the document retrieval
        batch_size: int
            Number of documents to retrieve per iteration
        include_vector: bool
            Include vectors in the search results
        sort: list
            Fields to sort by. For each field, sort by descending or ascending. If you are using descending by datetime, it will get the most recent ones.
        filters: list
            Query for filtering the search results
        """
        filters = [] if filters is None else filters
        sort = [] if sort is None else sort
        select_fields = [] if select_fields is None else select_fields

        return self._get_documents(
            dataset_id=self.dataset_id,
            number_of_documents=number_of_documents,
            filters=filters,
            cursor=cursor,
            batch_size=batch_size,
            sort=sort,
            select_fields=select_fields,
            include_vector=include_vector,
            include_cursor=include_cursor,
        )

    def get_metadata(self):
        """
        Store Metadata
        """
        return self.datasets.metadata(self.dataset_id)["results"]

    @property
    def metadata(self):
        """Get the metadata"""
        _metadata = self.get_metadata()
        self._metadata = Metadata(_metadata, self.credentials, self.dataset_id)
        return self._metadata

    def insert_metadata(self, metadata: dict):
        """Insert metadata"""
        results = self.datasets.post_metadata(self.dataset_id, metadata)
        if results == {}:
            print("✅ You have successfully inserted data.")
        else:
            return results

    def upsert_metadata(self, metadata: dict):
        """Upsert metadata."""
        original_metadata: dict = self.get_metadata()
        original_metadata.update(metadata)
        results = self.datasets.post_metadata(self.dataset_id, metadata)
        if results == {}:
            print("✅ You have successfully inserted metadata.")
        else:
            return results
        return self.insert_metadata(metadata)

    def chunk_dataset(
        self, select_fields: List = None, chunksize: int = 100, filters: list = None
    ):
        """Function for chunking a dataset"""
        docs = self.get_documents(
            number_of_documents=chunksize,
            include_cursor=True,
            filters=filters,
            select_fields=select_fields,
        )
        while len(docs["documents"]) > 0:
            yield docs["documents"]
            docs = self.get_documents(
                number_of_documents=chunksize,
                include_cursor=True,
                cursor=docs["cursor"],
                filters=filters,
            )
        return

    def list_vector_fields(self):
        """
        Returns list of valid vector fields in dataset
        Parameters
        ----------
        dataset_id : string
            Unique name of dataset

        Example
        ---------

        .. code-block::

            from relevanceai import Client
            client = Client()
            ds = client.Dataset("_mock_dataset_")
            ds.list_vector_fields()

        """
        schema = self.datasets.schema(self.dataset_id)
        return [
            k for k in schema.keys() if k.endswith("_vector_") and "_cluster_" not in k
        ]

    def list_cluster_aliases(self):
        raise NotImplementedError()

    def aggregate(
        self,
        groupby: Optional[list] = None,
        metrics: Optional[list] = None,
        filters: Optional[list] = None,
        page_size: int = 20,
        page: int = 1,
        asc: bool = False,
        flatten: bool = True,
        alias: str = "default",
    ):
        return self.services.aggregate.aggregate(
            dataset_id=self.dataset_id,
            metrics=groupby,
            groupby=metrics,
            page_size=page_size,
            page=page,
            asc=asc,
            flatten=flatten,
            alias=alias,
        )



from relevanceai.utils.base import _Base


class ClusterReportClient(_Base):
    def create(self, report: dict, name: str) -> dict:
        return self.make_http_request(
            "/reports/clusters/create",
            method="POST",
            parameters={"report": report, "name": name},
        )

    def get(self, report_id: str):
        return self.make_http_request(
            f"/reports/clusters/{report_id}/get",
            method="GET",
        )

    def update(self, report_id: str, report: dict, name: str):
        return self.make_http_request(
            f"/reports/clusters/{report_id}/update",
            method="POST",
            parameters={"report": report, "name": name},
        )

    def delete(self, report_id: str):
        return self.make_http_request(
            f"/reports/clusters/{report_id}/delete",
            method="POST",
        )

    def share(self):
        """Share all reports"""
        return self.make_http_request("/reports/clusters/share", method="POST")

    def private(self):
        return self.make_http_request("/reports/cluster/private", method="POST")

    def list(self):
        return self.make_http_request("/reports/clusters/list", method="GET")



import pytest

from sklearn.cluster import KMeans, MiniBatchKMeans

from relevanceai.dataset import Dataset

from .utils import VECTOR_FIELDS

ORIGINAL_ALIAS = "minibatchkmeans-3"


@pytest.fixture(scope="function")
def minibatch_clusterer(test_dataset: Dataset):
    alias = "minibatchkmeans-3"
    clusterer = test_dataset.cluster(
        model=MiniBatchKMeans(n_clusters=3),
        alias=alias,
        vector_fields=VECTOR_FIELDS,
    )
    yield clusterer
    clusterer.delete_centroids(test_dataset.dataset_id, VECTOR_FIELDS)


@pytest.fixture(scope="function")
def minibatch_subclusterer(test_dataset: Dataset):
    alias = "minibatchkmeans-4"
    clusterer = test_dataset.cluster(
        model=MiniBatchKMeans(n_clusters=4),
        alias=alias,
        vector_fields=VECTOR_FIELDS,
    )
    clusterer = test_dataset.cluster(
        model=MiniBatchKMeans(n_clusters=4),
        alias=alias,
        vector_fields=VECTOR_FIELDS,
        parent_alias=ORIGINAL_ALIAS,
    )
    yield test_dataset, alias
    clusterer.delete_centroids(test_dataset.dataset_id, VECTOR_FIELDS)


@pytest.fixture(scope="function")
def kmeans_subclusterer(test_dataset: Dataset):
    alias = "kmeans-4"
    clusterer = test_dataset.cluster(
        model=KMeans(n_clusters=4),
        alias=alias,
        vector_fields=VECTOR_FIELDS,
    )
    clusterer = test_dataset.cluster(
        model=KMeans(n_clusters=4),
        alias=alias,
        vector_fields=VECTOR_FIELDS,
        parent_alias=ORIGINAL_ALIAS,
    )
    yield test_dataset, alias
    clusterer.delete_centroids(test_dataset.dataset_id, VECTOR_FIELDS)



from tests.globals.documents.datetime_documents import *
from tests.globals.documents.error_documents import *
from tests.globals.documents.misc_documents import *
from tests.globals.documents.nested_documents import *
from tests.globals.documents.numpy_documents import *
from tests.globals.documents.pandas_documents import *
from tests.globals.documents.simple_documents import *
from tests.globals.documents.vector_documents import *



import pytest

from typing import Dict, List

from relevanceai import Client

from tests.globals.constants import generate_dataset_id


@pytest.fixture(scope="function")
def datetime_dataset(test_client: Client, datetime_documents: List[Dict]):
    test_dataset_id = generate_dataset_id()

    response = test_client._insert_documents(test_dataset_id, datetime_documents)

    yield response, len(datetime_documents)

    test_client.datasets.delete(test_dataset_id)



"""
ClusterOps class to run clustering. It is intended to be integrated with
models that inherit from `ClusterBase`.

You can run the ClusterOps as such:

.. code-block::

    from relevanceai import Client
    client = Client()
    df = client.Dataset("sample_dataset")

    from sklearn.cluster import KMeans
    model = KMeans(n_clusters=2)
    cluster_ops = client.ClusterOps(alias="kmeans_2", model=model)
    cluster_ops.fit_predict_update(df, vector_fields=["sample_vector_"])

You can view other examples of how to interact with this class here :ref:`integration`.

"""
import numpy as np

from tqdm.auto import tqdm

from typing import Union, Optional, Callable, Set, List, Dict, Any
from relevanceai.client.helpers import Credentials

from relevanceai.utils.decorators.analytics import track
from relevanceai.utils.decorators.version import beta

from relevanceai.operations.cluster.partial import PartialClusterOps
from relevanceai.operations.cluster.sub import SubClusterOps
from relevanceai.operations.cluster.groupby import ClusterGroupby, ClusterAgg
from relevanceai.operations.cluster.constants import METRIC_DESCRIPTION
from relevanceai.operations.cluster.base import (
    ClusterBase,
    CentroidClusterBase,
    BatchClusterBase,
)
from relevanceai.reports.cluster import ClusterReport
from relevanceai.constants.errors import NoDocumentsError, NoModelError


class ClusterOps(PartialClusterOps, SubClusterOps):
    """
    ClusterOps class allows users to set up any clustering model to fit on a Dataset.

    You can read about the other parameters here: https://scikit-learn.org/stable/modules/generated/sklearn.cluster.KMeans.html

    Example
    -----------

    .. code-block::

        from relevanceai import Client
        client = Client()
        df = client.Dataset("sample_dataset")

        from sklearn.cluster import KMeans
        model = KMeans(n_clusters=2)
        cluster_ops = client.ClusterOps(alias="kmeans_2", model=model)
        cluster_ops.fit_predict_update(df, vector_fields=["sample_vector_"])

    """

    def __init__(
        self,
        credentials: Credentials,
        model: Union[BatchClusterBase, ClusterBase, CentroidClusterBase, Any] = None,
        alias: Optional[str] = None,
        cluster_field: str = "_cluster_",
        parent_alias: str = None,
    ):
        self.credentials = credentials
        self.alias = alias  # type: ignore
        self.parent_alias = parent_alias
        self.cluster_field = cluster_field
        if model is None:
            raise NoModelError

        self.model = self._assign_model(model)

        self.verbose = True

        super().__init__(self.credentials)

    def __call__(self):
        self.groupby = ClusterGroupby(
            credentials=self.credentials,
            dataset_id=self.dataset_id,
            alias=self.alias,
            vector_fields=self.vector_fields,
        )
        self.agg = ClusterAgg(
            credentials=self.credentials,
            dataset_id=self.dataset_id,
            vector_fields=self.vector_fields,
            alias=self.alias,
        )

    @track
    def groupby(self):
        return ClusterGroupby(
            credentials=self.credentials,
            dataset_id=self.dataset_id,
            alias=self.alias,
            vector_fields=self.vector_fields,
        )

    @track
    def agg(self):
        self.groupby = ClusterGroupby(
            credentials=self.credentials,
            dataset_id=self.dataset_id,
            alias=self.alias,
            vector_fields=self.vector_fields,
        )
        return ClusterAgg(
            credentials=self.credentials,
            dataset_id=self.dataset_id,
            vector_fields=self.vector_fields,
            alias=self.alias,
        )

    @track
    def list_closest_to_center(
        self,
        vector_fields: Optional[List] = None,
        cluster_ids: Optional[List] = None,
        centroid_vector_fields: Optional[List] = None,
        select_fields: Optional[List] = None,
        approx: int = 0,
        sum_fields: bool = True,
        page_size: int = 1,
        page: int = 1,
        similarity_metric: str = "cosine",
        filters: Optional[List] = None,
        # facets: List = [],
        min_score: int = 0,
        include_vector: bool = False,
        include_count: bool = True,
    ):
        """
        List of documents closest from the centre.

        Parameters
        ----------

        cluster_ids: list
            Any of the cluster ids
        centroid_vector_fields: list
            Vector fields stored
        select_fields: list
            Fields to include in the search results, empty array/list means all fields
        approx: int
            Used for approximate search to speed up search. The higher the number, faster the search but potentially less accurate
        sum_fields: bool
            Whether to sum the multiple vectors similarity search score as 1 or seperate
        page_size: int
            Size of each page of results
        page: int
            Page of the results
        similarity_metric: string
            Similarity Metric, choose from ['cosine', 'l1', 'l2', 'dp']
        filters: list
            Query for filtering the search results
        facets: list
            Fields to include in the facets, if [] then all
        min_score: int
            Minimum score for similarity metric
        include_vectors: bool
            Include vectors in the search results
        include_count: bool
            Include the total count of results in the search results
        include_facets: bool
            Include facets in the search results

        Example
        --------------

        .. code-block::

            from relevanceai import Client
            client = Client()
            df = client.Dataset("sample_dataset_id")

            from sklearn.cluster import KMeans
            model = KMeans(n_clusters=2)
            cluster_ops = client.ClusterOps(alias="kmeans_2", model=model)
            cluster_ops.fit_predict_update(df, vector_fields=["sample_vector_"])

            cluster_ops.list_closest_to_center()

        """
        cluster_ids = [] if cluster_ids is None else cluster_ids
        centroid_vector_fields = (
            [] if centroid_vector_fields is None else centroid_vector_fields
        )
        select_fields = [] if select_fields is None else select_fields
        filters = [] if filters is None else filters

        return self.datasets.cluster.centroids.list_closest_to_center(
            dataset_id=self.dataset_id,
            vector_fields=self.vector_fields
            if vector_fields is None
            else vector_fields,
            alias=self.alias,
            cluster_ids=cluster_ids,
            centroid_vector_fields=centroid_vector_fields,
            select_fields=select_fields,
            approx=approx,
            sum_fields=sum_fields,
            page_size=page_size,
            page=page,
            similarity_metric=similarity_metric,
            filters=filters,
            min_score=min_score,
            include_vector=include_vector,
            include_count=include_count,
        )

    @track
    def aggregate(
        self,
        vector_fields: List[str] = None,
        metrics: Optional[list] = None,
        sort: Optional[list] = None,
        groupby: Optional[list] = None,
        filters: Optional[list] = None,
        page_size: int = 20,
        page: int = 1,
        asc: bool = False,
        flatten: bool = True,
    ):
        """
        Takes an aggregation query and gets the aggregate of each cluster in a collection. This helps you interpret each cluster and what is in them.
        It can only can be used after a vector field has been clustered. \n

        Aggregation/Groupby of a collection using an aggregation query. The aggregation query is a json body that follows the schema of:

        .. code-block::

            {
                "groupby" : [
                    {"name": <alias>, "field": <field in the collection>, "agg": "category"},
                    {"name": <alias>, "field": <another groupby field in the collection>, "agg": "numeric"}
                ],
                "metrics" : [
                    {"name": <alias>, "field": <numeric field in the collection>, "agg": "avg"}
                    {"name": <alias>, "field": <another numeric field in the collection>, "agg": "max"}
                ]
            }

        For example, one can use the following aggregations to group score based on region and player name.

        .. code-block::

            {
                "groupby" : [
                    {"name": "region", "field": "player_region", "agg": "category"},
                    {"name": "player_name", "field": "name", "agg": "category"}
                ],
                "metrics" : [
                    {"name": "average_score", "field": "final_score", "agg": "avg"},
                    {"name": "max_score", "field": "final_score", "agg": "max"},
                    {'name':'total_score','field':"final_score", 'agg':'sum'},
                    {'name':'average_deaths','field':"final_deaths", 'agg':'avg'},
                    {'name':'highest_deaths','field':"final_deaths", 'agg':'max'},
                ]
            }

        "groupby" is the fields you want to split the data into. These are the available groupby types:

            - category : groupby a field that is a category
            - numeric: groupby a field that is a numeric

        "metrics" is the fields and metrics you want to calculate in each of those, every aggregation includes a frequency metric. These are the available metric types:

            - "avg", "max", "min", "sum", "cardinality"

        The response returned has the following in descending order. \n

        If you want to return documents, specify a "group_size" parameter and a "select_fields" parameter if you want to limit the specific fields chosen. This looks as such:

            .. code-block::

                {
                    'groupby':[
                        {'name':'Manufacturer','field':'manufacturer','agg':'category',
                        'group_size': 10, 'select_fields': ["name"]},
                    ],
                    'metrics':[
                        {'name':'Price Average','field':'price','agg':'avg'},
                    ],
                }

                # ouptut example:
                {"title": {"title": "books", "frequency": 200, "documents": [{...}, {...}]}, {"title": "books", "frequency": 100, "documents": [{...}, {...}]}}

        For array-aggregations, you can add "agg": "array" into the aggregation query.

        Parameters
        ----------
        dataset_id : string
            Unique name of dataset
        metrics: list
            Fields and metrics you want to calculate
        groupby: list
            Fields you want to split the data into
        filters: list
            Query for filtering the search results
        page_size: int
            Size of each page of results
        page: int
            Page of the results
        asc: bool
            Whether to sort results by ascending or descending order
        flatten: bool
            Whether to flatten
        alias: string
            Alias used to name a vector field. Belongs in field_{alias} vector

        Parameters
        ----------
        metrics: list
            Fields and metrics you want to calculate
        groupby: list
            Fields you want to split the data into
        filters: list
            Query for filtering the search results
        page_size: int
            Size of each page of results
        page: int
            Page of the results
        asc: bool
            Whether to sort results by ascending or descending order
        flatten: bool
            Whether to flatten

        Example
        ---------

        .. code-block::

            from relevanceai import Client
            client = Client()
            df = client.Dataset("sample_dataset_id")

            from sklearn.cluster import KMeans
            model = KMeans(n_clusters=2)
            cluster_ops = client.ClusterOps(alias="kmeans_2", model=model)
            cluster_ops.fit_predict_update(df, vector_fields=["sample_vector_"])

            clusterer.aggregate(
                "sample_dataset_id",
                groupby=[{
                    "field": "title",
                    "agg": "wordcloud",
                }],
                vector_fields=['sample_vector_']
            )


        """
        metrics = [] if metrics is None else metrics
        sort = [] if sort is None else sort
        groupby = [] if groupby is None else groupby
        filters = [] if filters is None else filters

        return self.services.cluster.aggregate(
            dataset_id=self.dataset_id,
            vector_fields=self.vector_fields if not vector_fields else vector_fields,
            groupby=groupby,
            metrics=metrics,
            sort=sort,
            filters=filters,
            alias=self.alias,
            page_size=page_size,
            page=page,
            asc=asc,
            flatten=flatten,
        )

    def list_furthest_from_center(self, vector_fields: list = None):
        """
        List of documents furthest from the centre.

        Parameters
        ----------
        cluster_ids: list
            Any of the cluster ids
        select_fields: list
            Fields to include in the search results, empty array/list means all fields
        approx: int
            Used for approximate search to speed up search. The higher the number, faster the search but potentially less accurate
        sum_fields: bool
            Whether to sum the multiple vectors similarity search score as 1 or seperate
        page_size: int
            Size of each page of results
        page: int
            Page of the results
        similarity_metric: string
            Similarity Metric, choose from ['cosine', 'l1', 'l2', 'dp']
        filters: list
            Query for filtering the search results
        facets: list
            Fields to include in the facets, if [] then all
        min_score: int
            Minimum score for similarity metric
        include_vectors: bool
            Include vectors in the search results
        include_count: bool
            Include the total count of results in the search results
        include_facets: bool
            Include facets in the search results

        Example
        ---------
        .. code-block::

            from relevanceai import Client
            client = Client()
            df = client.Dataset("sample_dataset")

            from sklearn.cluster import KMeans
            model = KMeans(n_clusters=2)
            cluster_ops = client.ClusterOps(alias="kmeans_2", model=model)
            cluster_ops.fit_predict_update(df, vector_fields=["sample_vector_"])

            cluster_ops.list_furthest_from_center()

        """
        return self.datasets.cluster.centroids.list_furthest_from_center(
            dataset_id=self.dataset_id,
            vector_fields=self.vector_fields
            if vector_fields is None
            else vector_fields,
            alias=self.alias,
        )

    @track
    def insert_centroid_documents(self, centroid_documents: List[Dict]):
        """
        Insert the centroid documents

        Parameters
        ------------

        centroid_documents: List[Dict]
            Insert centroid documents
        dataset_id: str
            Dataset to insert

        Example
        ------------

        .. code-block::

            from relevanceai import Client
            client = Client()
            df = client.Dataset("sample_dataset")

            from sklearn.cluster import KMeans
            model = KMeans(n_clusters=2)
            cluster_ops = client.ClusterOps(alias="kmeans_2", model=model)
            cluster_ops.fit_predict_update(df, vector_fields=["sample_vector_"])

            centroids = cluster_ops.get_centroid_documents()
            cluster_ops.insert_centroid_documents(centroids)

        """

        results = self.services.cluster.centroids.insert(
            dataset_id=self.dataset_id,
            cluster_centers=centroid_documents,
            vector_fields=self.vector_fields,
            alias=self.alias,
        )
        return results

    @track
    def get_centroid_documents(self) -> List:
        """
        Get the centroid documents to store. This enables you to use `list_closest_to_center()`
        and `list_furthest_from_center`.

        .. code-block::

            {
                "_id": "document-id-1",
                "centroid_vector_": [0.23, 0.24, 0.23]
            }

        If multiple vector fields returns this:
        Returns multiple

        .. code-block::

            {
                "_id": "document-id-1",
                "blue_vector_": [0.12, 0.312, 0.42],
                "red_vector_": [0.23, 0.41, 0.3]
            }

        """
        if hasattr(self.model, "get_centroid_documents"):
            self.model.vector_fields = self.vector_fields
            return self.model.get_centroid_documents()
        self.centers = self.model.get_centers()

        if not hasattr(self, "vector_fields") or len(self.vector_fields) == 1:
            if isinstance(self.centers, np.ndarray):
                self.centers = self.centers.tolist()
            centroid_vector_field_name = self.vector_fields[0]
            return [
                {
                    "_id": self._label_cluster(i),
                    centroid_vector_field_name: self.centers[i],
                }
                for i in range(len(self.centers))
            ]
        # For one or more vectors, separate out the vector fields
        # centroid documents are created using multiple vector fields
        centroid_docs = []
        for i, c in enumerate(self.centers):
            centroid_doc = {"_id": self._label_cluster(i)}
            for j, vf in enumerate(self.vector_fields):
                centroid_doc[vf] = self.centers[i][vf]
            centroid_docs.append(centroid_doc.copy())
        return centroid_docs

    @property
    def centroids(self):
        """
        See your centroids if there are any.
        """
        return self.services.cluster.centroids.documents(
            self.dataset_id,
            vector_fields=self.vector_fields,
            alias=self.alias,
            page_size=10000,
            # cursor: str = None,
            include_vector=True,
        )

    @track
    def delete_centroids(self, dataset_id: str, vector_fields: List):
        """Delete the centroids after clustering."""
        # TODO: Fix delete centroids once its moved over to Node JS
        import requests

        base_url = self.config["api.base_url"]
        response = requests.post(
            base_url + "/services/cluster/centroids/delete",
            headers={"Authorization": self.project + ":" + self.api_key},
            params={
                "dataset_id": dataset_id,
                "vector_field": vector_fields,
                "alias": self.alias,
            },
        )
        return response.json()["status"]

    def fit_predict(
        self,
        vector_fields: List[str],
        filters: Optional[List[Dict]] = None,
        return_only_clusters: bool = True,
        include_report: bool = True,
        update: bool = True,
        inplace: bool = False,
    ):
        """
        Parameters
        ----------
        data: Union[str, Dataset, List[Dict]]
            Either a reference to a Relevance AI Dataset, be it its name
            (string) or the object itself (Dataset), or a list of documents
            (List[Dict]).

        vector_fields: List[str]
            The vector fields over which to fit the model.

        filters: List[Dict]
            A list of filters to enable for document retrieval. This only
            applies to a reference to a Relevance AI Dataset.

        return_only_clusters: bool
            An indicator that determines what is returned. If True, this
            function returns the clusters. Else, the function returns the
            original documents.

        include_report: bool
            An indictor that determines whether to include (True) a report and grade
            base on the mean silhouette score or not (False).

        update: bool
            An indicator that determines whether to update the documents
            that were part of the clustering process. This only applies to a
            reference to a Relevance AI Dataset.

        inplace: bool
            An indicator that determines whether the documents are edited
            inplace (True) or a copy is created and edited (False).

        Example
        -------

        .. code-block::

            from relevanceai import ClusterBase, Client

            client = Client()

            import random
            class CustomClusterModel(ClusterBase):
                def fit_predict(self, X):
                    cluster_labels = [random.randint(0, 100) for _ in range(len(X))]
                    return cluster_labels

            model = CustomClusterModel()

            df = client.Dataset("sample_dataset")
            clusterer = client.ClusterOps(alias="random_clustering", model=model)
            clusterer.fit_predict_update(df, vector_fields=["sample_vector_"])

        """
        filters = [] if filters is None else filters

        self.vector_fields = vector_fields
        # make sure to only get fields where vector fields exist
        filters.extend(
            [
                {
                    "field": f,
                    "filter_type": "exists",
                    "condition": "==",
                    "condition_value": " ",
                    "strict": "must_or",
                }
                for f in vector_fields
            ]
        )
        # load the documents
        self.logger.warning(
            "Retrieving documents... This can take a while if the dataset is large."
        )
        print("Retrieving all documents")
        documents = self._get_all_documents(
            dataset_id=self.dataset_id, filters=filters, select_fields=vector_fields
        )
        if len(documents) == 0:
            raise NoDocumentsError()

        vectors = self._get_vectors_from_documents(vector_fields, documents)

        # Label the clusters
        print("Fitting and predicting on all relevant documents")
        cluster_labels = self._label_clusters(self.model.fit_predict(vectors))

        if include_report:
            try:
                self._calculate_silhouette_grade(vectors, cluster_labels)
                report = ClusterReport(
                    X=vectors,
                    cluster_labels=cluster_labels,
                    num_clusters=len(cluster_labels),
                    model=self.model,
                )
                response = self.store_cluster_report(
                    report_name=self._get_cluster_field_name(self.alias),
                    report=report.internal_report,
                )
            except Exception as e:
                print(e)
                pass

        clustered_documents = self.set_cluster_labels_across_documents(
            cluster_labels,
            documents,
            inplace=inplace,
            return_only_clusters=return_only_clusters,
        )

        if update:
            # Updating the db
            print("Updating the database...")
            results = self._update_documents(
                self.dataset_id, clustered_documents, chunksize=10000
            )
            self.logger.info(results)

            # Update the centroid collection
            self.model.vector_fields = vector_fields

        self._insert_centroid_documents()
        print(
            "Build your clustering app here: "
            + f"https://cloud.relevance.ai/dataset/{self.dataset_id}/deploy/recent/cluster"
        )

        return clustered_documents

    @track
    def fit(
        self,
        dataset_id: str,
        vector_fields: List[str],
        filters: Optional[List] = None,
        include_report: bool = True,
        verbose: bool = True,
    ):
        """
        This function fits a cluster model onto a dataset. It sits under `fit`
        and is usually the only function that runs unless you have a get_centroid_documents
        function.

        Parameters
        ---------------
        dataset: Union[Dataset, str],
            The dataset object to fit it on
        vector_fields: list
            The vector fields to fit the model on
        filters: list
            The filters to run it on
        include_filters_for_vector_fields: bool
            If True, only cluster on those with the vector fields inside it.

        Example
        ---------

        .. code-block::

            from relevanceai import ClusterBase, Client
            client = Client()

            import random
            class CustomClusterModel(ClusterBase):
                def fit_predict(self, X):
                    cluster_labels = [random.randint(0, 100) for _ in range(len(X))]
                    return cluster_labels

            model = CustomClusterModel()

            clusterer = client.ClusterOps(model, alias="random_clustering")
            df = client.Dataset("sample_dataset")

            clusterer.fit_predict_update(df, vector_fields=["sample_vector_"])

        """
        self.dataset_id = dataset_id
        self.vector_fields = vector_fields

        filters = [] if filters is None else filters

        # load the documents
        self.logger.warning(
            "Retrieving documents... This can take a while if the dataset is large."
        )

        # make sure to only get fields where vector fields exist
        filters += [
            {
                "field": f,
                "filter_type": "exists",
                "condition": "==",
                "condition_value": " ",
            }
            for f in vector_fields
        ]
        if verbose:
            print("Retrieving all documents")
        fields_to_get = vector_fields.copy()
        if self.parent_alias:
            parent_field = self._get_cluster_field_name(self.parent_alias)
            fields_to_get.append(parent_field)

        # TODO: Figure out why this isn't cached
        docs = self._get_all_documents(
            dataset_id=dataset_id, filters=filters, select_fields=fields_to_get
        )

        if len(docs) == 0:
            raise NoDocumentsError()

        if verbose:
            print("Fitting and predicting on all documents")

        clustered_docs = self.fit_predict(
            vector_fields=vector_fields,
            return_only_clusters=True,
            inplace=False,
            include_report=include_report,
        )

        # Updating the db
        if verbose:
            print("Updating the database...")
        results = self._update_documents(dataset_id, clustered_docs, chunksize=10000)
        self.logger.info(results)

        # Update the centroid collection
        self.model.vector_fields = vector_fields

        self._insert_centroid_documents()

        if verbose:
            print(
                "Build your clustering app here: "
                + f"https://cloud.relevance.ai/dataset/{dataset_id}/deploy/recent/cluster"
            )

        return self

    def unique_cluster_ids(
        self,
        alias: str = None,
        minimum_cluster_size: int = 3,
        dataset_id: str = None,
        num_clusters: int = 1000,
    ):
        """
        We call facets on our data, which looks a little like this:

        {'results': {'_cluster_.sample_1_vector_.kmeans-3': [{'_cluster_.sample_1_vector_.kmeans-3': 'cluster-2',
                'frequency': 41,
                'value': 'cluster-2'},
            {'_cluster_.sample_1_vector_.kmeans-3': 'cluster-0',
                'frequency': 34,
                'value': 'cluster-0'},
            {'_cluster_.sample_1_vector_.kmeans-3': 'cluster-1',
                'frequency': 25,
                'value': 'cluster-1'}]}}
        """
        # Mainly to be used for subclustering
        # Get the cluster alias
        if dataset_id is None:
            self._check_for_dataset_id()
            dataset_id = self.dataset_id

        cluster_field = self._get_cluster_field_name(alias=alias)

        # currently the logic for facets is that when it runs out of pages
        # it just loops - therefore we need to store it in a simple hash
        # and then add them to a list
        all_cluster_ids: Set = set()

        while len(all_cluster_ids) < num_clusters:
            facet_results = self.datasets.facets(
                dataset_id=dataset_id,
                fields=[cluster_field],
                page_size=int(self.config["data.max_clusters"]),
                page=1,
                asc=True,
            )
            if "results" in facet_results:
                facet_results = facet_results["results"]
            if cluster_field not in facet_results:
                raise ValueError(
                    f"No clusters with alias `{alias}`. Please check the schema."
                )
            for facet in facet_results[cluster_field]:
                if facet["frequency"] > minimum_cluster_size:
                    curr_len = len(all_cluster_ids)
                    all_cluster_ids.add(facet[cluster_field])
                    new_len = len(all_cluster_ids)
                    if new_len == curr_len:
                        return list(all_cluster_ids)
        return list(all_cluster_ids)

    @track
    def fit_dataset(
        self,
        dataset,
        vector_fields,
        filters: Optional[list] = None,
        return_only_clusters: bool = True,
        inplace=False,
        verbose: bool = True,
    ):
        """ """
        filters = [] if filters is None else filters

        # load the documents
        self.logger.warning(
            "Retrieving documents... This can take a while if the dataset is large."
        )

        self._init_dataset(dataset)
        self.vector_fields = vector_fields

        # make sure to only get fields where vector fields exist
        filters += [
            {
                "field": f,
                "filter_type": "exists",
                "condition": "==",
                "condition_value": " ",
            }
            for f in vector_fields
        ]
        if verbose:
            print("Retrieving all documents")
        docs = self._get_all_documents(
            dataset_id=self.dataset_id, filters=filters, select_fields=vector_fields
        )

        if verbose:
            print("Fitting and predicting on all documents")
        return self.fit_predict(
            vector_fields,
            docs,
            return_only_clusters=True,
            inplace=inplace,
        )

    @track
    def predict_documents(
        self,
        vector_fields,
        documents,
        return_only_clusters: bool = True,
        inplace: bool = True,
    ):
        """
        Predict the documents
        """
        if len(vector_fields) > 1:
            raise ValueError("Currently do not suport more than 1 vector field.")
        vectors = self._get_vectors_from_documents(vector_fields, documents)
        cluster_labels = self.model.predict(vectors)
        cluster_labels = self._label_clusters(cluster_labels)
        return self.set_cluster_labels_across_documents(
            cluster_labels,
            documents,
            inplace=inplace,
            return_only_clusters=return_only_clusters,
        )

    @track
    def predict_update(
        self,
        dataset,
        vector_fields: Optional[List[str]] = None,
        chunksize: int = 20,
        verbose: bool = True,
    ):
        """
        Predict the dataset.

        Example
        ---------

        .. code-block::

            from relevanceai import Client
            client = Client()
            df = client.Dataset("sample_dataset")

            from sklearn.cluster import MiniBatchKMeans
            model = MiniBatchKMeans(n_clusters=2)
            cluster_ops = client.ClusterOps(alias="minibatchkmeans_2", model=model)

            cluster_ops.partial_fit_dataset(df)
            cluster_ops.predict_dataset(df)

        """
        if not vector_fields:
            vector_fields = self.vector_fields

        all_responses = {
            "inserted": 0,
            "failed_documents": [],
            "failed_documents_detailed": [],
        }

        filters = [
            {
                "field": f,
                "filter_type": "exists",
                "condition": "==",
                "condition_value": " ",
            }
            for f in vector_fields
        ]

        for c in self._chunk_dataset(
            dataset, self.vector_fields, chunksize=chunksize, filters=filters
        ):
            cluster_predictions = self.predict_documents(
                vector_fields=vector_fields, documents=c
            )
            response = self.dataset._update_documents(
                dataset_id=self.dataset_id, documents=cluster_predictions
            )
            for k, v in response.items():
                if isinstance(all_responses[k], int):
                    all_responses["inserted"] += v
                elif isinstance(all_responses[k], list):
                    all_responses[k] += v
        if verbose:
            print(
                "Build your clustering app here: "
                + f"https://cloud.relevance.ai/dataset/{self.dataset_id}/deploy/recent/cluster"
            )
        return all_responses

    def set_cluster_labels_across_documents(
        self,
        cluster_labels: list,
        documents: List[Dict],
        inplace: bool = True,
        return_only_clusters: bool = True,
    ):
        """
        Utility function to allow users to set cluster labels

        Parameters
        ------------
        cluster_labels: List[str, int]
            A list of integers of string. If it is an integer - it will automatically add a 'cluster-' prefix
            to help avoid incorrect data type parsing. You can override this behavior by setting clusters
            as strings.
        documents: List[dict]
            When the documents are in
        inplace: bool
            If True, then the clusters are set in place.
        return_only_clusters: bool
            If True, then the return_only_clusters will return documents with just the cluster field and ID.
            This can be helpful when you want to upsert quickly without having to re-insert the entire document.

        Example
        -----------

        .. code-block::

            labels = list(range(10))
            documents = [{"_id": str(x)} for x in range(10)]
            cluster_ops.set_cluster_labels_across_documents(labels, documents)

        """
        if inplace:
            self._set_cluster_labels_across_documents(cluster_labels, documents)
            if return_only_clusters:
                return [
                    {"_id": d.get("_id"), self.cluster_field: d.get(self.cluster_field)}
                    for d in documents
                ]
            return documents

        # useful if you want to upload as quickly as possible
        new_documents = documents.copy()

        self._set_cluster_labels_across_documents(cluster_labels, new_documents)
        if return_only_clusters:
            return [
                {"_id": d.get("_id"), self.cluster_field: d.get(self.cluster_field)}
                for d in new_documents
            ]
        return new_documents

    @property
    def metadata(self):
        """
        If metadata is none, retrieves metadata about a dataset. notably description, data source, etc
        Otherwise, you can store the metadata about your cluster here.

        Example
        ----------

        .. code-block::

            from relevanceai import Client
            client = Client()
            df = client.Dataset("_github_repo_vectorai")
            from relevanceai.ops.clusterops.kmeans_clusterer import KMeansModel

            model = KMeansModel()
            kmeans = client.ClusterOps(model, alias="kmeans_sample")
            kmeans.fit(df, vector_fields=["sample_1_vector_"])
            kmeans.metadata
            # {"k": 10}

        """
        return self.services.cluster.centroids.metadata(
            dataset_id=self.dataset_id,
            vector_fields=self.vector_fields,
            alias=self.alias,
        )

    @metadata.setter
    def metadata(self, metadata: dict):
        """
        If metadata is none, retrieves metadata about a dataset. notably description, data source, etc
        Otherwise, you can store the metadata about your cluster here.

        Parameters
        ----------
        metadata: dict
           If None, it will retrieve the metadata, otherwise
           it will overwrite the metadata of the cluster

        Example
        ----------

        .. code-block::


            from relevanceai import Client
            client = Client()
            df = client.Dataset("sample_dataset")

            from sklearn.cluster import KMeans
            model = KMeans(n_clusters=2)
            cluster_ops = client.ClusterOps(alias="kmeans_2", model=model)

            cluster_ops.fit(df, vector_fields=["sample_1_vector_"])
            cluster_ops.metadata
            # {"k": 10}

        """
        return self.services.cluster.centroids.metadata(
            dataset_id=self.dataset_id,
            vector_fields=self.vector_fields,
            alias=self.alias,
            metadata=metadata,
        )

    @track
    def evaluate(
        self,
        ground_truth_column: Union[str, None] = None,
        metric: Union[str, Callable] = "euclidean",
        random_state: Union[int, None] = None,
        average_score: bool = False,
    ):
        """
        Evaluate your clusters using the silhouette score, or if true labels are provided, using completeness, randomness and homogeniety as well

        Parameters
        ----------
        ground_truth_index : str
            the index of the true label of each sample in dataset

        metric : str or Callable
            A string referencing suportted distance functions, or custom method for calculating the distance between 2 vectors

        random_state : int, default None
            for reproducability

        average_score : bool
            a boolean that determines whether to average the evaluation metrics in a new a score. only applicable if ground_truth_index is not None

        Example
        ----------

        .. code-block::


            from relevanceai import Client
            client = Client()
            df = client.Dataset("sample_dataset")

            from sklearn.cluster import KMeans
            model = KMeans(n_clusters=2)
            cluster_ops = client.ClusterOps(alias="kmeans_2", model=model)

            cluster_ops.fit(df, vector_fields=["sample_vector_"])

            cluster_ops.evaluate()
            cluster_ops.evaluate("truth_column")
        """

        from sklearn.metrics import (
            adjusted_rand_score,
            completeness_score,
            silhouette_score,
            homogeneity_score,
        )

        vector_field = self.vector_fields[0]
        cluster_field = self.cluster_field
        alias = self.alias

        samples = self._get_all_documents(self.dataset_id, include_vector=True)

        vectors = [sample[vector_field] for sample in samples]
        pred_labels = [sample[cluster_field][vector_field][alias] for sample in samples]

        def norm(value, min, max):
            return (value - min) / (max - min)

        s_score = silhouette_score(
            vectors, pred_labels, metric=metric, random_state=random_state
        )

        stats = {}
        stats["silhouette"] = {
            "score": s_score,
            "description": METRIC_DESCRIPTION["silhouette"],
        }

        if (
            ground_truth_column
        ):  # if the ground truth column was provided, we can run these cluster evaluation measures

            ground_truth_labels = {  # convert cluster1 etc. to respective label found in ground truth column
                key: value[0][ground_truth_column]
                for key, value in self.list_closest_to_center(page_size=1).items()
                if value
            }
            pred_labels = (
                [  # get the predicted clusters and return their respective true labels
                    ground_truth_labels[sample[cluster_field][vector_field][alias]]
                    for sample in samples
                ]
            )
            true_labels = [
                sample[ground_truth_column] for sample in samples
            ]  # same for actual labels

            ar_score = adjusted_rand_score(true_labels, pred_labels)  # compute metric
            c_score = completeness_score(true_labels, pred_labels)  # compute metric
            h_score = homogeneity_score(true_labels, pred_labels)  # compute metric

            stats["random"] = {
                "score": ar_score,
                "description": METRIC_DESCRIPTION["random"],
            }
            stats["completeness"] = {
                "score": c_score,
                "description": METRIC_DESCRIPTION["completeness"],
            }
            stats["homogeneity"] = {
                "score": h_score,
                "description": METRIC_DESCRIPTION["homogeneity"],
            }

            if average_score:  # If the user wants an average as well
                ns_score = norm(s_score, min=-1, max=1)
                nar_score = norm(ar_score, min=-1, max=1)

                average_score = (ns_score + nar_score + c_score + h_score) / 4

                stats["average"] = {
                    "score": average_score,
                    "description": METRIC_DESCRIPTION["average"],
                }

        return stats

    @beta
    @track
    def store_cluster_report(
        self, report_name: str, report: dict, verbose: bool = True
    ):
        """

        Store the cluster data.

        .. code-block::

            from relevanceai import Client
            client = Client()
            client.store_cluster_report("sample", {"value": 3})

        """
        response: dict = self.reports.clusters.create(
            name=report_name, report=self.json_encoder(report)
        )
        if verbose:
            print(
                f"You can now access your report at https://cloud.relevance.ai/report/cluster/{self.region}/{response['_id']}"
            )
        return response

    @track
    def internal_report(self, verbose: bool = True):
        """
        Get a report on your clusters.

        Example
        ---------

        This is what is returned on an `auto_cluster` method.

        .. code-block::

            from relevanceai.package_utils.datasets import mock_documents
            docs = mock_documents(10)
            df = client.Dataset('sample')
            df.upsert_documents(docs)
            cluster_ops = df.auto_cluster('kmeans-2', ['sample_1_vector_'])
            cluster_ops.internal_report()

        This is what is returned on an `internal_report` method.

        .. code-block::

            from relevanceai import Client
            # client = Client()
            from relevanceai.package_utils.datasets import mock_documents
            ds = client.Dataset("sample")
            # Creates 100 sample documents
            documents = mock_documents(100)
            ds.upsert_documents(documents)

            from sklearn.cluster import KMeans
            model = KMeans(n_clusters=10)
            clusterer = client.ClusterOps(alias="not-auto-kmeans-10", model=model)

            clusterer.fit_predict_update(dataset=ds, vector_fields=["sample_1_vector_"])
            cluster_ops.internal_report()

        """
        if isinstance(self.vector_fields, list) and len(self.vector_fields) > 1:
            raise ValueError(
                "We currently do not support more than 1 vector field when reporting."
            )
        from relevanceai.reports.cluster import ClusterReport

        # X is all the vectors
        cluster_field_name = self._get_cluster_field_name()
        all_docs = self._get_all_documents(
            self.dataset_id, select_fields=self.vector_fields + [cluster_field_name]
        )
        cluster_labels = self.get_field_across_documents(cluster_field_name, all_docs)
        self.number_of_clusters = len(set(cluster_labels))
        self._report = ClusterReport(
            self.get_field_across_documents(self.vector_fields[0], all_docs),
            cluster_labels=self.get_field_across_documents(
                cluster_field_name, all_docs
            ),
            model=self.model,
            num_clusters=self.number_of_clusters,
        )
        cluster_response = self.store_cluster_report(
            report_name=cluster_field_name,
            report=self.json_encoder(self._report.internal_report),
            verbose=verbose,
        )

        return self._report.internal_report

    report = internal_report

    def operate(self, field: str, func: Callable, output_field: Optional[str] = None):
        """
        Run an function per cluster.

        Parameters
        ------------

        field: str
            The field to operate on
        func: Callable
            The function to run on all the values once they are received.
            It should take in a list of values
        output_field: Optional[str]
            Outputs for every document in the cluster

        Example
        ---------

        .. code-block::

            import numpy as np
            def vector_mean(vectors):
                return np.mean(vectors, axis=0)
            cluster_centroids = cluster_ops.operate(
                field="review_sentence_answer_use_vector_",
                func=vector_mean
            )

        """
        # Run a function on each cluster
        output: Dict = {}
        cluster_ids = self.unique_cluster_ids()
        for cluster_id in tqdm(cluster_ids):
            self._operate(cluster_id, field, output, func)
        if output_field is not None:
            self.update_documents_within_clusters(output, output_field)
        return output

    def create_centroids(
        self, vector_fields: List[str], operation: Optional[Callable] = None
    ):
        """
        Create centroids if there are none. The default operation is to take the centroid
        of each vector.
        An alternative function can be provided provided.

        Example of the operation in question for mean:

        .. code-block::

            def vector_mean(vectors):
                return np.mean(vectors, axis=0)

            clusterops.create_centroids(["sample_vector_"], operation=vector_mean)

        """
        if len(vector_fields) > 1:
            raise ValueError("currently do not support more than 1 vector field.")
        vector_field: str = vector_fields[0]

        def vector_mean(vectors):
            return np.mean(vectors, axis=0)

        if operation == "mean" or operation is None:
            operation = vector_mean

        cluster_centroids = self.operate(field=vector_field, func=operation)  # type: ignore
        centroid_docs = []
        for k, v in cluster_centroids.items():
            centroid_docs.append({"_id": str(k), vector_field: v.tolist()})
        return self.insert_centroid_documents(centroid_docs)

    def update_documents_within_cluster(self, cluster_id: str, update: dict):
        """
        Update all the documents within a cluster
        """
        cluster_filter = self._get_filter_for_cluster(cluster_id)
        result = self.datasets.documents.update_where(
            dataset_id=self.dataset_id, update=update, filters=cluster_filter
        )
        return result

    def update_documents_within_clusters(
        self, cluster_id_and_updates: dict, output_field: str = None
    ):
        """
        Takes the cluster ids and updates and updates them accordingly

        Example
        ---------

        .. code-block::

            # Let us take the operator results
            operator_results = clusterops.operate(...)
            clusterops.update_documents_within_clusters(
                operator_results, output_field="centroid_vector_"
            )
        """
        results = []
        for cluster_id, update_value in cluster_id_and_updates.items():
            if type(update_value) != dict:
                if output_field is None:
                    raise ValueError(
                        """
                        As update value is not a dictionary, you will need to
                        specify `output_field=`.
                    """
                    )
                update: dict = {}
                self.set_field(output_field, update, update_value)
            else:
                update = update_value

            cluster_filter = self._get_filter_for_cluster(cluster_id)
            result = self.datasets.documents.update_where(
                dataset_id=self.dataset_id, update=update, filters=cluster_filter
            )
            results.append(result)
        return results

    # def show(
    #     self,
    #     dataset_id: str,
    #     field: str,
    #     vector_field: str,
    #     alias: str,
    #     preview_num: int = 10,
    #     is_image_field: bool = False,
    # ):
    #     """
    #     Shows the values of a clustering.

    #     dataset_id: str
    #         The dataset ID of with the clustering of interest.

    #     field: str
    #         The field whose values are of interest.

    #     vector_field: str
    #         The vector field that was used for the clustering.

    #     alias: str
    #         The alias of the clustering.

    #     preview_num: int
    #         The maximum number of values to preview. If a cluster has fewer
    #         values than preview_num, the cluster list will be extended by
    #         the appropriate number of np.nans to make up the difference.

    #     is_image_field: bool
    #         If set to True, will show the values as the images rather than
    #         links.
    #     """
    #     if type(preview_num) is not int and preview_num <= 0:
    #         raise TypeError(
    #             f"Please provide a valid non-zero integer for {preview_num}."
    #         )

    #     if not vector_field.startswith(field):
    #         raise ValueError(f"{vector_field} must be a vector of {field}.")

    #     # Since there is no gaurantee self.dataset_id will be not None, it is
    #     # safer to force the user to specify the dataset_id
    #     ds = Dataset(
    #         self.project, self.api_key, dataset_id, self.firebase_uid, fields=[]
    #     )
    #     schema = ds.schema

    #     if field not in schema:
    #         raise ValueError(f"{field} does not exist")

    #     cluster_field = ".".join(["_cluster_", vector_field])
    #     if cluster_field not in schema:
    #         raise ValueError(f"{vector_field} has not been clustered yet.")

    #     alias_field = ".".join([cluster_field, alias])
    #     if alias_field not in schema:
    #         raise ValueError(f"A clustering with alias {alias} has not been made yet.")

    #     cluster_values = ds.to_pandas_dataframe(
    #         select_fields=[alias_field, field], show_progress_bar=False
    #     )
    #     # converts {"vector_field": {"alias": "cluster-k"}} to "cluster-k"
    #     cluster_values["_cluster_"] = cluster_values["_cluster_"].apply(
    #         lambda cluster: cluster[vector_field][alias]
    #     )

    #     cluster_groups = dict(list(cluster_values.groupby("_cluster_")[field]))
    #     clusters = {}
    #     for cluster, values in cluster_groups.items():
    #         length = min(len(values), preview_num)
    #         # It is faster to index a list than a pandas Series
    #         clusters[cluster] = values.tolist()[:length]

    #         # If the number of values in the cluster is fewer than the desired
    #         # preview number, insert np.nan to make up the difference. Note
    #         # that the range of a negative number is not a range.
    #         difference = preview_num - len(clusters[cluster])
    #         clusters[cluster].extend([np.nan for _ in range(difference)])

    #     # return pd.DataFrame(clusters)
    #     return _ClusterOpsShow(pd.DataFrame(clusters), is_image_field)



from typing import Dict, Optional

from relevanceai.client.helpers import Credentials
from relevanceai.utils.base import _Base


class DeployableClient(_Base):
    def __init__(self, credentials: Credentials):
        super().__init__(credentials)

    def create(self, dataset_id: str, configuration: Optional[Dict] = None):
        """
        Create a private deployable from an existing dataset.
        TODO: explain what a deployable is

        Parameters
        ------------
        dataset_id: string
            Unique name of dataset
        configuration: None | dict
            A configuration specification

        Returns
        ---------
        On success (200): {
            deployable_id, dataset_id, project_id, api_key, configuration
        }
        On failure (422): {
            loc, msg, type
        }
        """
        if configuration is None:
            configuration = {}

        return self.make_http_request(
            endpoint="/deployables/create",
            method="POST",
            parameters={"dataset_id": dataset_id, "configuration": configuration},
        )

    def share(self, deployable_id: str):
        """
        Share a private deployable.
        The response should be {"status": "success"} or {"status": "failed"}

        Parameters
        ----------
        deployable_id: string
            Unique name of deployable

        Returns
        -------
        On success (200): {
            configuration, status, message
        }
        On failure (422): {
            loc, msg, type
        }
        """
        return self.make_http_request(
            endpoint=f"/deployables/{deployable_id}/share", method="POST"
        )

    def unshare(self, deployable_id: str):
        """
        Unshare a shared deployable, making it private.
        The response should be {"status": "success"} or {"status": "failed"}

        Parameters
        ----------
        deployable_id: string
            Unique name of deployable

        Returns
        -------
        On success (200): {
            configuration, status, message
        }
        On failure (422): {
            loc, msg, type
        }
        """
        return self.make_http_request(
            endpoint=f"/deployables/{deployable_id}/private", method="POST"
        )

    def update(
        self,
        deployable_id: str,
        dataset_id: str,
        configuration: Optional[Dict] = None,
        overwrite: bool = True,
        upsert: bool = True,
    ):
        """
        Update a specified deployable.

        Parameters
        ----------
        deployable_id: string
            The deployable configuration
        dataset_id: string
            Unique name of dataset.
        configuration: None | dict
            The deployable configuration
        overwrite: Boolean
            Whether to overwrite document if it exists
        upsert: Boolean
            If True, adds new fields. This is prioritized over overwrite.

        Returns
        -------
        On success (200): {
            configuration, status, message
        }
        On failure (422): {
            loc, msg, type
        }
        """
        return self.make_http_request(
            endpoint=f"/deployables/{deployable_id}/update",
            method="POST",
            parameters={
                "dataset_id": dataset_id,
                "configuration": configuration,
                "overwrite": overwrite,
                "upsert": upsert,
            },
        )

    def get(self, deployable_id: str):
        """
        Get a specified deployable.

        Parameters
        ----------
        deployable_id: string
            The deployable configuration

        Returns
        -------
        On success (200): {
            project_id, dataset_id, api_key, configuration
        }
        On failure (422): {
            loc, msg, type
        }
        """
        return self.make_http_request(
            endpoint=f"/deployables/{deployable_id}/get", method="GET"
        )

    def delete(self, deployable_id: str):
        """
        Delete a specified deployable.

        Parameters
        ----------
        deployable_id: string
            The deployable configuration

        Returns
        -------
        On success (200): {
            status, message
        }
        On failure (422): {
            loc, msg, type
        }
        """
        return self.make_http_request(
            endpoint=f"/deployables/delete",
            method="POST",
            parameters={
                "id": deployable_id,
            },
        )

    def list(self):
        """
        List all deployables.

        Parameters
        ----------
            None

        Returns
        -------
        On success (200): {
            deployables, count
        }
        On failure (422): {
            loc, msg, type
        }
        """
        return self.make_http_request(endpoint="/deployables/list", method="GET")

    def url(self, deployable_id: str, dataset_id: str, application: str) -> str:
        """
        Generates the deployable URL.

        Parameters
        ----------
        deployable_id: str
            ID of the deployable
        dataset_id: str
            ID of the dataset
        application: str
            The type of deployable application
        """
        url = "https://cloud.relevance.ai/dataset/{}/deploy/{}/{}/{}/{}"
        return url.format(
            dataset_id, self.project, application, self.api_key, deployable_id
        )



import numpy as np
import warnings
from typing import Any, Dict, List, Optional, Tuple, Union

from relevanceai.client.helpers import Credentials
from relevanceai.constants import CLUSTER_APP_LINK, Warning
from relevanceai._api import APIClient


class ClusterOps(APIClient):
    def __init__(
        self,
        credentials: Credentials,
        model: Union[str, Any],
        vector_fields: Optional[List[str]] = None,
        alias: Optional[str] = None,
        n_clusters: Optional[int] = None,
        cluster_config: Optional[Dict[str, Any]] = None,
        outlier_value: int = -1,
        outlier_label: str = "outlier",
        **kwargs,
    ):
        """
        ClusterOps object

        Parameters
        -------------

        model: Union[str, Any]
            The string of clustering algorithm, class of clustering algorithm or custom clustering class.
            If custom, the model must contain the method for fit_predict and must output a numpy array for labels

        vector_fields: Optional[List[str]] = None
            A list of vector_fields to cluster over

        alias: Optional[str] = None
            An alias to be given for clustering.
            If no alias is provided, alias becomes the cluster model name hypen n_clusters e.g. (kmeans-5, birch-8)

        n_clusters: Optional[int] = None
            Number of clusters to find, should the argument be required of the algorithm.
            If None, n_clusters defaults to 8 for sklearn models ONLY.

        cluster_config: Optional[Dict[str, Any]] = None
            Hyperparameters for the clustering model.
            See the documentation for supported clustering models for list of hyperparameters

        outlier_value: int = -1
            For unsupervised clustering models like OPTICS etc. outliers are given the integer label -1

        outlier_label: str = "outlier"
            When viewing the cluster app dashboard, outliers will be prefixed with outlier_label

        """
        self.vector_field = None if vector_fields is None else vector_fields[0]

        self.config = {} if cluster_config is None else cluster_config  # type: ignore
        if n_clusters is not None:
            self.config["n_clusters"] = n_clusters  # type: ignore

        self.model_name = None
        self.model = self._get_model(model)

        if "package" not in self.__dict__:
            self.package = self._get_package(self.model)

        if hasattr(self.model, "n_clusters"):
            self.n_clusters = self.model.n_clusters
        else:
            self.n_clusters = n_clusters

        self.alias = self._get_alias(alias)
        self.outlier_value = outlier_value
        self.outlier_label = outlier_label

        super().__init__(credentials, **kwargs)

    def __call__(self, dataset_id: str, vector_fields: List[str]) -> None:
        self.forward(dataset_id=dataset_id, vector_fields=vector_fields)

    def _get_alias(self, alias: Any) -> str:
        # Auto-generates alias here
        if alias is None:
            if hasattr(self.model, "n_clusters"):
                n_clusters = (
                    self.n_clusters
                    if self.n_clusters is not None
                    else self.model.n_clusters
                )
                alias = f"{self.model_name}-{n_clusters}"

            elif hasattr(self.model, "k"):
                n_clusters = (
                    self.n_clusters if self.n_clusters is not None else self.model.k
                )
                alias = f"{self.model_name}-{n_clusters}"

            else:
                alias = self.model_name

            Warning.MISSING_ALIAS.format(alias=alias)
        return alias.lower()

    def _get_package(self, model):
        model_name = str(model.__class__)
        if "function" in model_name:
            model_name = str(model.__name__)

        if "sklearn" in model_name:
            package = "sklearn"

        elif "faiss" in model_name:
            package = "faiss"

        elif "hdbscan" in model_name:
            package = "hdbscan"

        elif "community_detection" in model_name:
            package = "sentence-transformers"

        return package

    def _get_model(self, model):
        if isinstance(model, str):
            model = model.lower().replace(" ", "").replace("_", "")
            self.model_name = model
        else:
            self.model_name = str(model.__class__).split(".")[-1].split("'>")[0]
            self.package = "custom"
            return model

        if isinstance(model, str):
            if model == "affinitypropagation":
                from sklearn.cluster import AffinityPropagation

                model = AffinityPropagation(**self.config)

            elif model == "agglomerativeclustering":
                from sklearn.cluster import AgglomerativeClustering

                model = AgglomerativeClustering(**self.config)

            elif model == "birch":
                from sklearn.cluster import Birch

                model = Birch(**self.config)

            elif model == "dbscan":
                from sklearn.cluster import DBSCAN

                model = DBSCAN(**self.config)

            elif model == "optics":
                from sklearn.cluster import OPTICS

                model = OPTICS(**self.config)

            elif model == "kmeans":
                from sklearn.cluster import KMeans

                model = KMeans(**self.config)

            elif model == "featureagglomeration":
                from sklearn.cluster import FeatureAgglomeration

                model = FeatureAgglomeration(**self.config)

            elif model == "meanshift":
                from sklearn.cluster import MeanShift

                model = MeanShift(**self.config)

            elif model == "minibatchkmeans":
                from sklearn.cluster import MiniBatchKMeans

                model = MiniBatchKMeans(**self.config)

            elif model == "spectralclustering":
                from sklearn.cluster import SpectralClustering

                model = SpectralClustering(**self.config)

            elif model == "spectralbiclustering":
                from sklearn.cluster import SpectralBiclustering

                model = SpectralBiclustering(**self.config)

            elif model == "spectralcoclustering":
                from sklearn.cluster import SpectralCoclustering

                model = SpectralCoclustering(**self.config)

            elif model == "hdbscan":
                from hdbscan import HDBSCAN

                model = HDBSCAN(**self.config)

            elif model == "community_detection":
                # TODO: this is a callable (?)
                from sentence_transformers.util import community_detection

                model = community_detection

            elif "faiss" in model:
                from faiss import Kmeans

                model = Kmeans(**self.config)

        else:
            # TODO: this needs to be referenced from relevance.constants.errors
            raise ValueError("ModelNotSupported")

        return model

    def _format_labels(self, labels: np.ndarray) -> List[str]:
        labels = labels.flatten().tolist()
        cluster_labels = [
            f"cluster-{str(label)}"
            if label == self.outlier_value
            else self.outlier_label
            for label in labels
        ]
        return cluster_labels

    def _get_centroid_documents(
        self, vectors: np.ndarray, labels: List[str]
    ) -> List[Dict[str, Any]]:
        centroid_documents = []

        centroids: Dict[str, Any] = {}
        for label, vector in zip(labels, vectors.tolist()):
            if label not in centroids:
                centroids[label] = []
            centroids[label].append(vector)

        for centroid, vectors in centroids.items():
            centroid_vector = np.array(vectors).mean(0).tolist()
            centroid_document = dict(
                _id=centroid,
                centroid_vector=centroid_vector,
            )
            centroid_documents.append(centroid_document)

        return centroid_documents

    def _insert_centroids(
        self,
        dataset_id: str,
        vector_field: str,
        centroid_documents: List[Dict[str, Any]],
    ) -> None:
        self.services.cluster.centroids.insert(
            dataset_id=dataset_id,
            cluster_centers=centroid_documents,
            vector_fields=[vector_field],
            alias=self.alias,
        )

    def _fit_predict(
        self, documents: List[Dict[str, Any]], vector_field: str
    ) -> Tuple[List[Dict[str, Any]], List[Dict[str, Any]]]:
        vectors = np.array(
            [
                self.get_field(vector_field, document)
                for document in documents
                if self.is_field(vector_field, document)
            ]
        )

        if self.package == "sklearn":
            labels = self.model.fit_predict(vectors)

        elif self.package == "hdbscan":
            self.model.fit(vectors)
            labels = self.model.labels_.reshape(-1, 1)

        elif self.package == "faiss":
            vectors = vectors.astype("float32")
            self.model.train(vectors)
            labels = self.model.assign(vectors)[1]

        elif self.package == "sentencetransformers":
            # TODO: make teh config here better
            labels = self.model(vectors)
            labels = np.array(labels)

        elif self.package == "custom":
            labels = self.model.fit_predict(vectors)
            if not isinstance(labels, np.ndarray):
                raise ValueError("Custom model must output np.ndarray for labels")

            if labels.shape[0] != vectors.shape[0]:
                raise ValueError("incorrect shape")

        labels = self._format_labels(labels)

        cluster_field = f"_cluster_.{vector_field}.{self.alias}"
        self.set_field_across_documents(
            field=cluster_field, values=labels, docs=documents
        )

        centroid_documents = self._get_centroid_documents(vectors, labels)

        return centroid_documents, documents

    def _print_app_link(self):
        link = CLUSTER_APP_LINK.format(self.dataset_id)
        print(link)

    def forward(
        self,
        dataset_id: Union[str, Any],
        vector_fields: List[str],
        show_progress_bar: bool = True,
    ) -> None:
        if not isinstance(dataset_id, str):
            if hasattr(dataset_id, "dataset_id"):
                dataset_id = dataset_id.dataset_id

        self.dataset_id = dataset_id
        vector_field = vector_fields[0]
        self.vector_field = vector_field

        # get all documents
        documents = self._get_all_documents(
            dataset_id=dataset_id,
            select_fields=vector_fields,
            show_progress_bar=show_progress_bar,
            include_vector=True,
        )

        # fit model, predict and label all documents
        centroid_documents, labelled_documents = self._fit_predict(
            documents=documents,
            vector_field=vector_field,
        )

        # update all documents
        self._update_documents(
            dataset_id=dataset_id,
            documents=labelled_documents,
            show_progress_bar=show_progress_bar,
        )

        # insert centroids
        self._insert_centroids(
            dataset_id=dataset_id,
            vector_field=vector_field,
            centroid_documents=centroid_documents,
        )

        # link back to dashboard
        self._print_app_link()

    def closest(
        self,
        dataset_id: Optional[str] = None,
        vector_field: Optional[str] = None,
        alias: Optional[str] = None,
    ):
        dataset_id = self.dataset_id if dataset_id is None else dataset_id
        vector_field = self.vector_field if vector_field is None else vector_field
        alias = self.alias if alias is None else alias

        return self.services.cluster.centroids.list_closest_to_center(
            dataset_id=dataset_id,
            vector_fields=[vector_field],
            alias=alias,
        )

    def furthest(
        self,
        dataset_id: Optional[str] = None,
        vector_field: Optional[str] = None,
        alias: Optional[str] = None,
    ):
        """
        List furthest from cluster centers
        """
        dataset_id_ = self.dataset_id if dataset_id is None else dataset_id
        vector_field_ = self.vector_field if vector_field is None else vector_field
        alias_ = self.alias if alias is None else alias

        return self.services.cluster.centroids.list_furthest_from_center(
            dataset_id=dataset_id_,
            vector_fields=[vector_field_],  # type: ignore
            alias=alias_,
        )

    # Convenience functions
    list_closest = closest
    list_furthest = furthest



import numpy as np
import pandas as pd

from relevanceai.utils import make_id


def pandas_document():
    document = {
        "sample_1_pandas": pd.DataFrame(
            np.random.randint(0, 20, size=(20, 4)), columns=list("ABCD")
        ),
        "sample_2_pandas": pd.DataFrame(
            np.random.randint(0, 10, size=(10, 4)), columns=list("ABCD")
        ),
        "sample_3_pandas": pd.DataFrame(
            np.array([[1, 2, np.nan], [4, np.nan, 6], [7, 8, 9]]),
            columns=["a", "b", "c"],
        ),
    }
    document["_id"] = make_id(document)
    return document



from relevanceai.client.helpers import Credentials
from relevanceai.utils.base import _Base
from relevanceai._api.endpoints.datasets.cluster.centroids import CentroidsClient


class ClusterClient(_Base):
    def __init__(self, credentials: Credentials):
        self.centroids = CentroidsClient(credentials)
        super().__init__(credentials)



"""
Create workflow diagrams.
"""


def create_diagram(workflows):
    try:
        import graphviz
    except ModuleNotFoundError:
        raise ModuleNotFoundError(
            "Install graphviz following instructions here: https://graphviz.org/download/"
        )

    dot = graphviz.Digraph(
        engine="dot",
        name="Experimentation Workflow",
        graph_attr={
            "splines": "ortho",
        },
        node_attr={
            "shape": "rect",
        },
        format="png",
    )

    nodes = []
    for i, workflow in enumerate(workflows):
        input_field = workflow["input_field"] + "-field"
        output_field = workflow["output_field"] + "-field"
        dot.node(
            input_field, workflow["input_field"], style="filled", color="lightgreen"
        )  # doctest: +NO_EXE
        dot.node(
            output_field,
            workflow["output_field"],
            style="filled",
            bgcolor="green",
            color="lightblue",
        )
        dot.edge(
            input_field,
            output_field,
            label=workflow["workflow_alias"],
            color="green",
            shape="box",
        )
    return dot



"""Multithreading Module
"""
import math
from concurrent.futures import ProcessPoolExecutor, ThreadPoolExecutor, as_completed
from typing import Callable

from relevanceai.utils.progress_bar import NullProgressBar, progress_bar


def chunk(iterables, n=20):
    return [iterables[i : i + n] for i in range(0, int(len(iterables)), int(n))]


def multithread(
    func, iterables, max_workers=8, chunksize=20, show_progress_bar: bool = False
):
    # with progress_bar(total=int(len(iterables) / chunksize), show_progress_bar=show_progress_bar) as pbar:
    with ThreadPoolExecutor(max_workers=max_workers) as executor:
        # Create trackers
        progress_tracker = progress_bar(
            range(math.ceil(len(iterables) / chunksize)),
            show_progress_bar=show_progress_bar,
        )
        progress_iterator = iter(progress_tracker)

        futures = [executor.submit(func, it) for it in chunk(iterables, chunksize)]
        results = []
        for future in as_completed(futures):
            results.append(future.result())
            if progress_tracker is not None:
                next(progress_iterator)
            if show_progress_bar is True:
                progress_tracker.update(1)
        return results


def multiprocess(
    func,
    iterables,
    max_workers=8,
    chunksize=20,
    post_func_hook: Callable = None,
    show_progress_bar: bool = False,
    process_args: tuple = (),
):
    # with progress_bar(total=int(len(iterables) / chunksize),
    #     show_progress_bar=show_progress_bar) as pbar:
    with ProcessPoolExecutor(max_workers=max_workers) as executor:
        # Create trackers
        progress_tracker = progress_bar(
            range(math.ceil(len(iterables) / chunksize)),
            show_progress_bar=show_progress_bar,
        )
        progress_iterator = iter(progress_tracker)

        futures = [
            executor.submit(func, it, process_args)
            for it in chunk(iterables, chunksize)
        ]
        results = []
        for future in as_completed(futures):
            if post_func_hook:
                results.append(post_func_hook(future.result()))
            else:
                results.append(future.result())
            if progress_tracker is not None:
                next(progress_iterator)
            if show_progress_bar is True:
                progress_tracker.update(1)
        return results



import math
import warnings
from relevanceai.constants.warning import Warning

import pandas as pd
import numpy as np

from typing import Dict, List, Union, Callable, Optional

from tqdm import tqdm

from relevanceai.client.helpers import Credentials
from relevanceai.constants import MAX_CACHESIZE
from relevanceai.utils.cache import lru_cache
from relevanceai.utils.decorators.analytics import track
from relevanceai._api import APIClient


class Series(APIClient):
    """
    Dataset Series Object
    -----------------------------

    A wrapper class for being able to vectorize documents over field

    Parameters
    ----------
    project : str
        Project name on RelevanceAI
    api_key : str
        API key for RelevanceAI
    dataset_id : str
        Data type for the output Series. If not specified, this will be
        inferred from `data`.
        See the :ref:`user guide <basics.dtypes>` for more usages.
    field : str
        The name of the field with the Dataset.

    Examples
    --------
    Assuming the following code has been executed:

    .. code-block::

        from relevanceai import client
        relevanceai.package_utils.datasets import get_dummy_ecommerce_dataset

        documents = get_dummy_ecommerce_dataset()
        client = Client()

        df = client.Dataset('ecommerce')
        df.create()
        df.insert_documents(documents)

    Retrieve a Series from your dataset

    .. code-block::

        product_images = df['product_image'] # A Series object of every every product image url in dataset

    """

    def __init__(
        self,
        credentials: Credentials,
        dataset_id: str,
        field: str,
        image_fields: Optional[List[str]] = None,
        audio_fields: Optional[List[str]] = None,
        highlight_fields: Optional[Dict[str, List]] = None,
        text_fields: Optional[List[str]] = None,
    ):
        super().__init__(credentials)
        self.dataset_id = dataset_id
        self.field = field
        self.image_fields = [] if image_fields is None else image_fields
        self.audio_fields = [] if audio_fields is None else audio_fields
        self.highlight_fields = {} if highlight_fields is None else highlight_fields
        self.text_fields = [] if text_fields is None else text_fields

    def _repr_html_(self):
        if "_vector_" in self.field:
            include_vector = True
        else:
            include_vector = False

        documents = self._get_documents(
            dataset_id=self.dataset_id,
            select_fields=[self.field],
            include_vector=include_vector,
        )
        if include_vector:
            warnings.warn(Warning.MISSING_RELEVANCE_NOTEBOOK)
            return pd.json_normalize(documents).set_index("_id")._repr_html_()

        try:
            return self._show_json(documents, return_html=True)
        except Exception as e:
            warnings.warn(Warning.MISSING_RELEVANCE_NOTEBOOK + str(e))
            return pd.json_normalize(documents).set_index("_id")._repr_html_()

    @track
    def list_aliases(self):
        fields = self._format_select_fields()
        fields = [field for field in fields if field != "_cluster_"]
        return pd.DataFrame(fields, columns=["_cluster_"])._repr_html_()

    def _format_select_fields(self):
        fields = [
            field
            for field in self.datasets.schema(self.dataset_id)
            if "_cluster_" in field
        ]
        return fields

    def _show_json(self, documents, **kw):
        from jsonshower import show_json

        if not self.text_fields:
            text_fields = pd.json_normalize(documents).columns.tolist()
        else:
            text_fields = self.text_fields

        return show_json(
            documents,
            image_fields=self.image_fields,
            audio_fields=self.audio_fields,
            highlight_fields=self.highlight_fields,
            text_fields=text_fields,
            **kw,
        )

    @track
    def sample(
        self,
        n: int = 1,
        frac: float = None,
        filters: Optional[list] = None,
        random_state: int = 0,
        include_vector: bool = True,
        output_format="pandas",
    ):
        """
        Return a random sample of items from a dataset.

        Parameters
        ----------
        n : int
            Number of items to return. Cannot be used with frac.
        frac: float
            Fraction of items to return. Cannot be used with n.
        filters: list
            Query for filtering the search results
        random_state: int
            Random Seed for retrieving random documents.

        Example
        -------
        .. code-block::

            from relevanceai import client

            client = Client()

            df = client.Dataset(dataset_id)
            df.sample(n=3)

        """
        filters = [] if filters is None else filters

        select_fields = [self.field]

        if frac and n:
            raise ValueError("Only one of n or frac can be provided")

        if frac:
            if frac > 1 or frac < 0:
                raise ValueError("Fraction must be between 0 and 1")
            n = math.ceil(
                self.get_number_of_documents(self.dataset_id, filters=filters) * frac
            )

        documents = self.datasets.documents.get_where(
            dataset_id=self.dataset_id,
            filters=filters,
            page_size=n,
            random_state=random_state,
            is_random=True,
            select_fields=select_fields,
            include_vector=include_vector,
        )["documents"]

        if output_format == "json":
            return documents
        elif output_format == "pandas":
            return pd.DataFrame.from_dict(documents, orient="records")
        else:
            raise ValueError("Incorrect output format")

    head = sample

    @lru_cache(maxsize=MAX_CACHESIZE)
    @track
    def all(
        self,
        chunksize: int = 1000,
        filters: Optional[List] = None,
        sort: Optional[List] = None,
        include_vector: bool = True,
        show_progress_bar: bool = True,
    ):
        filters = [] if filters is None else filters
        sort = [] if sort is None else sort

        select_fields = [self.field] if isinstance(self.field, str) else self.field
        return self._get_all_documents(
            dataset_id=self.dataset_id,
            chunksize=chunksize,
            filters=filters,
            sort=sort,
            select_fields=select_fields,
            include_vector=include_vector,
            show_progress_bar=show_progress_bar,
        )

    @track
    def apply(
        self,
        func: Callable,
        output_field: str,
        filters: list = [],
        axis: int = 0,
    ):
        """
        Apply a function along an axis of the DataFrame.

        Objects passed to the function are Series objects whose index is either the DataFrame’s index (axis=0) or the DataFrame’s columns (axis=1). By default (result_type=None), the final return type is inferred from the return type of the applied function. Otherwise, it depends on the result_type argument.


        .. note::
            We recommend using the bulk_apply functionality
            if you are looking to have faster processing.

        Parameters
        --------------
        func: function
            Function to apply to each document

        axis: int
            Axis along which the function is applied.
            - 9 or 'index': apply function to each column
            - 1 or 'columns': apply function to each row

        output_field: str
            The field from which to output

        Example
        ---------------
        .. code-block::

            from relevanceai import Client

            client = Client()

            dataset_id = "sample_dataset_id"
            df = client.Dataset(dataset_id)

            df["sample_1_label"].apply(lambda x: x + 3, output_field="output_field")

        """
        if axis == 1:
            raise ValueError("We do not support column-wise operations!")

        def bulk_fn(documents):
            for d in tqdm(documents):
                try:
                    if self.is_field(self.field, d):
                        self.set_field(
                            output_field, d, func(self.get_field(self.field, d))
                        )
                except Exception as e:
                    continue
            return documents

        return self.pull_update_push_async(
            self.dataset_id, bulk_fn, select_fields=[self.field], filters=filters
        )

    @track
    def bulk_apply(
        self,
        bulk_func: Callable,
        retrieve_chunksize: int = 100,
        filters: Optional[list] = None,
        select_fields: Optional[list] = None,
        show_progress_bar: bool = True,
        use_json_encoder: bool = True,
    ):
        """
        Apply a bulk function along an axis of the DataFrame.

        Parameters
        ------------
        bulk_func: function
            Function to apply to a bunch of documents at a time
        retrieve_chunksize: int
            The number of documents that are received from the original collection with each loop iteration.
        max_workers: int
            The number of processors you want to parallelize with
        max_error: int
            How many failed uploads before the function breaks
        json_encoder : bool
            Whether to automatically convert documents to json encodable format
        axis: int
            Axis along which the function is applied.
            - 9 or 'index': apply function to each column
            - 1 or 'columns': apply function to each row

        Example
        ---------
        .. code-block::

            from relevanceai import Client
            client = Client()

            df = client.Dataset("sample_dataset_id")

            def update_documents(documents):
                for d in documents:
                    d["value"] = 10
                return documents

            df.bulk_apply(update_documents)
        """
        filters = [] if filters is None else filters
        select_fields = [] if select_fields is None else select_fields

        return self.pull_update_push_async(
            self.dataset_id,
            bulk_func,
            retrieve_chunk_size=retrieve_chunksize,
            filters=filters,
            select_fields=select_fields,
            show_progress_bar=show_progress_bar,
            use_json_encoder=use_json_encoder,
        )

    @track
    def numpy(self) -> np.ndarray:
        """
        Iterates over all documents in dataset and returns all numeric values in a numpy array.

        Parameters
        ---------
        None

        Returns
        -------
        vectors: np.ndarray
            an array/matrix of all numeric values selected

        Example
        ---------------
        .. code-block::

            from relevanceai import Client

            client = Client()

            dataset_id = "sample_dataset_id"
            df = client.Dataset(dataset_id)

            field = "sample_field"
            arr = df[field].numpy()
        """
        documents = self._get_all_documents(self.dataset_id, select_fields=[self.field])
        vectors = self.get_field_across_documents(self.field, documents)
        vectors = np.array(vectors)
        return vectors

    @track
    def value_counts(
        self,
        normalize: bool = False,
        ascending: bool = False,
        sort: bool = False,
        bins: Optional[int] = None,
    ):
        """
        Return a Series containing counts of unique values (or values with in a range if bins is set).

        Parameters
        ----------
        normalize : bool, default False
            If True then the object returned will contain the relative frequencies of the unique values.
        ascending : bool, default False
            Sort in ascending order.
        bins : int, optional
            Groups categories into 'bins'. These bins are good for representing groups within continuous series

        Returns
        ----------
        Series

        Example
        ---------------
        .. code-block::

            from relevanceai import Client

            client = Client()

            dataset_id = "sample_dataset_id"
            df = client.Dataset(dataset_id)

            field = "sample_field"
            value_counts_df = df[field].value_counts()
        """
        schema = self.datasets.schema(self.dataset_id)
        dtype = schema[self.field]

        if dtype == "numeric":
            agg_type = dtype
        else:
            agg_type = "category"

        groupby_query = [{"name": self.field, "field": self.field, "agg": agg_type}]
        aggregation = self.services.aggregate.aggregate(
            self.dataset_id,
            groupby=groupby_query,
            page_size=10000,
            asc=ascending,
        )

        total = self.get_number_of_documents(dataset_id=self.dataset_id)
        aggregation = pd.DataFrame(aggregation)

        if normalize:
            aggregation["frequency"] /= total

        if bins is not None:
            vals = []
            for agg in [[agg[0]] * int(agg[1]) for agg in aggregation.values]:
                vals += agg

            vals = pd.cut(vals, bins)

            categories = [
                "({}, {}]".format(interval.left, interval.right) for interval in vals
            ]
            unique_categories = list(set(categories))

            if sort:
                categories = sorted(
                    categories, key=lambda x: float(x.split(",")[0][1:])
                )

            aggregation = pd.DataFrame(
                [categories.count(cat) for cat in unique_categories],
                index=unique_categories,
            )
            aggregation.columns = ["Frequency"]

        return aggregation

    def __getitem__(self, loc: Union[int, str]):
        """
        Indexs a value with a series, usually to get a specific sample from a column in your dataset

        Parameters
        ----------
        loc : int or str, preferably a str
            if int, this operates exactly as indexing a regular python list
            if str, this will be a string corresponding to the _id of the document

        Returns
        ----------
        A single document

        Example
        ---------------
        .. code-block::

            from relevanceai import Client

            client = Client()

            dataset_id = "sample_dataset_id"
            df = client.Dataset(dataset_id)

            field = "sample_field"
            id = "sample_id"
            index = 56

            document = df[field][id]
            document = df[field][index]
        """
        if isinstance(loc, int):
            warnings.warn(Warning.INDEX_STRING)
            return self.get_documents(loc + 1, select_fields=[self.field])[loc][
                self.field
            ]
        elif isinstance(loc, str):
            return self.datasets.documents.get(self.dataset_id, loc)[self.field]
        raise TypeError("Incorrect data type! Must be a string or an integer")

    @lru_cache(maxsize=MAX_CACHESIZE)
    def _get_pandas_series(self):
        documents = self._get_all_documents(
            dataset_id=self.dataset_id,
            select_fields=[self.field],
            include_vector=False,
            show_progress_bar=True,
        )

        try:
            df = pd.DataFrame(documents)
            df.set_index("_id", inplace=True)
            return df.squeeze()
        except KeyError:
            raise Exception("No documents found")

    def __getattr__(self, attr):
        if hasattr(pd.Series, attr):
            series = self._get_pandas_series()
            try:
                return getattr(series, attr)
            except SyntaxError:
                raise AttributeError(f"'{attr}' is an invalid attribute")
        raise AttributeError(f"'{attr}' is an invalid attribute")

    def __add__(self, other):
        schema = self.datasets.schema(self.dataset_id)
        if schema[self.field] != "numeric":
            raise ValueError(f"{self.field} is not a numeric type")
        if schema[other.field] != "numeric":
            raise ValueError(f"{other.field} is not a numeric type")
        if other.field not in schema:
            raise ValueError(f"{other.field} must be an attribute of {self.dataset_id}")

        return self._get_pandas_series() + other._get_pandas_series()



# -*- coding: utf-8 -*-

from abc import abstractmethod
import pandas as pd
import numpy as np
import json

from doc_utils import DocUtils

from typing import List, Union, Dict, Any, Tuple, Optional
from typing_extensions import Literal
from relevanceai.client.helpers import Credentials

from relevanceai.utils.base import _Base
from relevanceai.utils.logger import LoguruLogger
from relevanceai.operations.cluster.constants import (
    DIM_REDUCTION,
    DIM_REDUCTION_DEFAULT_ARGS,
)


class DimReductionBase(LoguruLogger, DocUtils):
    def __call__(self, *args, **kwargs):
        return self.fit_predict(*args, **kwargs)

    # @abstractmethod
    def fit_transform(self, *args, **kw) -> np.ndarray:
        raise NotImplementedError

    def fit(self, *args, **kw):
        raise NotImplementedError

    def transform(self, *args, **kw):
        raise NotImplementedError

    def transform_documents(self, vector_field: str, documents: List[Dict]):
        vectors = self.get_field_across_documents(vector_field, documents)
        return self.transform(documents)

    def fit_documents(self, vector_field: str, documents: List[Dict]):
        vectors = self.get_field_across_documents(
            vector_field, documents, missing_treatment="skip"
        )
        return self.fit(vectors)

    def get_dr_vector_field_name(self, vector_field: str, alias: str):
        return ".".join(
            [
                "_dr_",
                alias,
                vector_field,
            ]
        )

    def fit_transform_documents(
        self,
        vector_field: str,
        documents: List[Dict],
        alias: str,
        exclude_original_vectors: bool = True,
        dims: int = 3,
    ):
        documents = list(self.filter_docs_for_fields([vector_field], documents))
        vectors = np.array(
            self.get_field_across_documents(
                vector_field, documents, missing_treatment="skip"
            )
        )
        dr_vectors = self.fit_transform(vectors, dims=dims)
        dr_vector_field_name = self.get_dr_vector_field_name(vector_field, alias)
        self.set_field_across_documents(dr_vector_field_name, dr_vectors, documents)
        if exclude_original_vectors:
            dr_docs = self.subset_documents(["_id", dr_vector_field_name], documents)
        return dr_docs


class PCA(DimReductionBase):
    def fit(self, vectors: np.ndarray, dims: int = 3, *args, **kw):
        from sklearn.decomposition import PCA as SKLEARN_PCA

        pca = SKLEARN_PCA(n_components=min(dims, vectors.shape[1]))
        return pca.fit(vectors)

    def fit_transform(
        self,
        vectors: np.ndarray,
        dr_args: Optional[Dict[Any, Any]] = DIM_REDUCTION_DEFAULT_ARGS["pca"],
        dims: int = 3,
    ) -> np.ndarray:
        from sklearn.decomposition import PCA as SKLEARN_PCA

        self.logger.debug(f"{dr_args}")
        vector_length = len(vectors[0])
        pca = SKLEARN_PCA(n_components=min(dims, vector_length), **dr_args)
        return pca.fit_transform(vectors)


class TSNE(DimReductionBase):
    def fit_transform(
        self,
        vectors: np.ndarray,
        dr_args: Optional[Dict[Any, Any]] = DIM_REDUCTION_DEFAULT_ARGS["tsne"],
        dims: int = 3,
    ) -> np.ndarray:
        from sklearn.decomposition import PCA
        from sklearn.manifold import TSNE

        pca = PCA(n_components=min(10, vectors.shape[1]))
        data_pca = pca.fit_transform(vectors)
        self.logger.debug(f"{dr_args}")
        tsne = TSNE(n_components=dims, **dr_args)
        return tsne.fit_transform(data_pca)


class UMAP(DimReductionBase):
    def fit_transform(
        self,
        vectors: np.ndarray,
        dr_args: Optional[Dict[Any, Any]] = DIM_REDUCTION_DEFAULT_ARGS["umap"],
        dims: int = 3,
    ) -> np.ndarray:
        try:
            from umap import UMAP
        except ModuleNotFoundError as e:
            raise ModuleNotFoundError(
                f"{e}\nInstall umap\n \
                pip install -U relevanceai[umap]"
            )
        self.logger.debug(f"{dr_args}")
        umap = UMAP(n_components=dims, **dr_args)
        return umap.fit_transform(vectors)


class Ivis(DimReductionBase):
    def fit_transform(
        self,
        vectors: np.ndarray,
        dr_args: Optional[Dict[Any, Any]] = DIM_REDUCTION_DEFAULT_ARGS["ivis"],
        dims: int = 3,
    ) -> np.ndarray:
        try:
            from ivis import Ivis
        except ModuleNotFoundError as e:
            raise ModuleNotFoundError(
                f"{e}\nInstall ivis\n \
                CPU: pip install -U ivis-cpu\n \
                GPU: pip install -U ivis-gpu"
            )
        self.logger.debug(f"{dr_args}")
        ivis = Ivis(embedding_dims=dims, **dr_args)
        if ivis.batch_size > vectors.shape[0]:
            ivis.batch_size = vectors.shape[0]
        vectors_dr = ivis.fit(vectors).transform(vectors)
        return vectors_dr


class DimReduction(_Base, DimReductionBase):
    def __init__(self, credentials: Credentials):
        super().__init__(credentials)

    @staticmethod
    def dim_reduce(
        vectors: np.ndarray,
        dr: Union[DIM_REDUCTION, DimReductionBase],
        dr_args: Union[None, dict],
        dims: Literal[2, 3],
    ) -> np.ndarray:
        """
        Dimensionality reduction
        """
        if isinstance(dr, str):
            if dr_args is None:
                dr_args = DIM_REDUCTION_DEFAULT_ARGS[dr]
            if dr == "pca":
                return PCA().fit_transform(vectors=vectors, dr_args=dr_args, dims=dims)
            elif dr == "tsne":
                return TSNE().fit_transform(vectors=vectors, dr_args=dr_args, dims=dims)
            elif dr == "umap":
                return UMAP().fit_transform(vectors=vectors, dr_args=dr_args, dims=dims)
            elif dr == "ivis":
                return Ivis().fit_transform(vectors=vectors, dr_args=dr_args, dims=dims)
            raise ValueError("not suppported")
        elif isinstance(dr, DimReductionBase):
            return dr().fit_transform(vectors=vectors, dr_args=dr_args, dims=dims)
        return np.array([])



"""
Test cloning and sending a dataset. Simple tests to check functionality.
In truth, should be handled API side.
"""
import pytest
import time

from typing import Dict, List

from relevanceai import Client


@pytest.mark.skip(reason="Node has not implemented yet.")
def test_sending_dataset(test_client: Client):

    DATASET_ID = "_sample_dataset_id_"
    docs = [{"_id": "10", "value": 10}, {"_id": "1000", "value": 30}]
    test_client.insert_documents(DATASET_ID, docs)
    test_client.send_dataset(
        dataset_id=DATASET_ID,
        receiver_project=test_client.project,
        receiver_api_key=test_client.api_key,
    )
    test_client.delete_dataset(DATASET_ID)
    assert True


@pytest.mark.skip(reason="Node has not implemented yet.")
def test_cloning_dataset(test_client: Client):
    DATASET_ID = "_sample_dataset_id_"
    NEW_DATASET_ID = DATASET_ID + "-2"
    docs = [{"_id": "10", "value": 10}, {"_id": "1000", "value": 30}]
    test_client.insert_documents(DATASET_ID, docs)
    test_client.clone_dataset(
        source_dataset_id=DATASET_ID, new_dataset_id=NEW_DATASET_ID
    )
    time.sleep(10)
    test_client.delete_dataset(NEW_DATASET_ID)
    assert True



from relevanceai.client.helpers import Credentials
from relevanceai._api import APIClient


class Metadata(APIClient):
    """Metadata object"""

    def __init__(self, metadata: dict, credentials: Credentials, dataset_id: str):
        self._metadata = metadata
        self.dataset_id = dataset_id
        super().__init__(credentials)

    def __repr__(self):
        return str(self.to_dict())

    def __contains__(self, m):
        return m in self.to_dict().keys()

    def insert_metadata(self, metadata: dict, verbose: bool = False):
        """Insert metadata"""
        results = self.datasets.post_metadata(self.dataset_id, metadata)
        if results == {}:
            return
        else:
            return results

    def upsert_metadata(self, metadata: dict, verbose: bool = False):
        """Upsert metadata."""
        original_metadata: dict = self.datasets.metadata(self.dataset_id)
        original_metadata.update(metadata)

        # TODO: how do you fire and forget this
        return self.insert_metadata(metadata)

    def __getitem__(self, key):
        return self.get_field(key, self._metadata)

    def __setitem__(self, key, value):
        self.set_field(key, self._metadata, value)
        self.upsert_metadata(self._metadata)

    def to_dict(self):
        return self._metadata
